# -*- coding: utf-8 -*-
import numpy as np
import torch

import pandas as pd
from logging import getLogger
import os

from accelbrainbase.extractabledata.unlabeled_csv_extractor import UnlabeledCSVExtractor
from accelbrainbase.iteratabledata.labeled_image_iterator import LabeledImageIterator as _LabeledImageIterator
from accelbrainbase.noiseable_data import NoiseableData


class UnlabeledTHotCSVIterator(_LabeledImageIterator):
    '''
    Iterator that draws from CSV files and generates `mxnet.ndarray` of unlabeled samples.
    '''

    def __init__(
        self,
        unlabeled_csv_extractor,
        train_csv_path_list,
        test_csv_path_list,
        epochs=300,
        batch_size=20,
        seq_len=10,
        norm_mode="z_score",
        scale=1.0,
        noiseable_data=None,
        ctx="cpu"
    ):
        '''
        Init.

        Args:
            image_extractor:                is-a `ImageExtractor`.
            train_csv_path_list:            `list` of `str` of path to CSV file in training.
            test_csv_path_list:             `list` of `str` of path to CSV file in test.

            epochs:                         `int` of epochs of Mini-batch.
            batch_size:                     `int` of batch size of Mini-batch.
            seq_len:                        `int` of the length of series. 
            norm_mode:                      How to normalize pixel values of images.
                                            - `z_score`: Z-Score normalization.
                                            - `min_max`: Min-max normalization.
                                            - others : This class will not normalize the data.

            scale:                          `float` of scaling factor for data.
            noiseable_data:                 is-a `NoiseableData` for Denoising Auto-Encoders.
        '''
        if isinstance(unlabeled_csv_extractor, UnlabeledCSVExtractor) is False:
            raise TypeError("The type of `unlabeled_csv_extractor` must be `UnlabeledCSVExtractor`.")
        if noiseable_data is not None and isinstance(noiseable_data, NoiseableData) is False:
            raise TypeError("The type of `noiseable_data` must be `NoiseableData`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        self.__unlabeled_csv_extractor = unlabeled_csv_extractor

        self.__train_csv_path_list = train_csv_path_list
        self.__test_csv_path_list = test_csv_path_list

        dataset_size = 0
        for file_key in range(len(self.__train_csv_path_list)):
            train_observed_arr = self.__unlabeled_csv_extractor.extract(
                self.__train_csv_path_list[file_key]
            )
            dataset_size = dataset_size + train_observed_arr.shape[0]

        iter_n = int(epochs * max(dataset_size / batch_size, 1))

        self.iter_n = iter_n
        self.epochs = epochs
        self.batch_size = batch_size
        self.seq_len = seq_len
        self.norm_mode = norm_mode
        self.scale = scale
        self.__noiseable_data = noiseable_data

        self.__ctx = ctx

    def generate_learned_samples(self):
        '''
        Draw and generate data.

        Returns:
            `Tuple` data. The shape is ...
            - `mxnet.ndarray` of observed data points in training.
            - `mxnet.ndarray` of supervised data in training.
            - `mxnet.ndarray` of observed data points in test.
            - `mxnet.ndarray` of supervised data in test.
        '''
        for _ in range(self.iter_n):
            training_batch_arr, test_batch_arr = None, None

            for i in range(self.batch_size):
                file_key = np.random.randint(low=0, high=len(self.__train_csv_path_list))
                train_observed_arr = self.__unlabeled_csv_extractor.extract(
                    self.__train_csv_path_list[file_key]
                )
                test_file_key = np.random.randint(low=0, high=len(self.__test_csv_path_list))
                test_observed_arr = self.__unlabeled_csv_extractor.extract(
                    self.__test_csv_path_list[test_file_key]
                )

                train_observed_arr = np.identity(
                    1 + int(train_observed_arr.max() + (train_observed_arr.min() * -1))
                )[
                    (train_observed_arr.reshape(train_observed_arr.shape[0], -1) + (train_observed_arr.min() * -1)).astype(int)
                ]

                test_observed_arr = np.identity(
                    1 + int(test_observed_arr.max() + (test_observed_arr.min() * -1))
                )[
                    (test_observed_arr.reshape(test_observed_arr.shape[0], -1) + (test_observed_arr.min() * -1)).astype(int)
                ]

                start_row = np.random.randint(low=0, high=train_observed_arr.shape[0] - self.seq_len)
                test_start_row = np.random.randint(low=0, high=test_observed_arr.shape[0] - self.seq_len)

                train_observed_arr = train_observed_arr[start_row:start_row+self.seq_len]
                test_observed_arr = test_observed_arr[test_start_row:test_start_row+self.seq_len]

                if training_batch_arr is None:
                    training_batch_arr = torch.from_numpy(train_observed_arr)
                    training_batch_arr = training_batch_arr.to(self.__ctx).float()
                    training_batch_arr = torch.unsqueeze(training_batch_arr, dim=0)
                else:
                    train_observed_arr = torch.from_numpy(train_observed_arr)
                    train_observed_arr = train_observed_arr.to(self.__ctx).float()
                    training_batch_arr = torch.cat(
                        training_batch_arr,
                        torch.unsqueeze(
                            train_observed_arr,
                            dim=0
                        ),
                        dim=0
                    )

                test_observed_arr = torch.from_numpy(test_observed_arr)
                test_observed_arr = test_observed_arr.to(self.__ctx).float()
                if test_batch_arr is None:
                    test_batch_arr = torch.unsqueeze(
                        test_observed_arr,
                        dim=0
                    )
                else:
                    test_batch_arr = torch.cat(
                        (
                            test_batch_arr,
                            torch.unsqueeze(
                                test_observed_arr,
                                dim=0
                            )
                        ),
                        dim=0
                    )

            training_batch_arr = self.pre_normalize(training_batch_arr)
            test_batch_arr = self.pre_normalize(test_batch_arr)

            if self.__noiseable_data is not None:
                training_batch_arr = self.__noiseable_data.noise(training_batch_arr)

            yield training_batch_arr, training_batch_arr, test_batch_arr, test_batch_arr

    def generate_inferenced_samples(self):
        '''
        Draw and generate data.
        The targets will be drawn from all image file sorted in ascending order by file name.

        Returns:
            `Tuple` data. The shape is ...
            - `None`.
            - `None`.
            - `mxnet.ndarray` of observed data points in test.
            - file path.
        '''
        test_batch_arr = None

        for test_file_key in range(len(self.__test_csv_path_list)):
            test_observed_arr = self.__unlabeled_csv_extractor.extract(
                self.__test_csv_path_list[test_file_key]
            )
            test_observed_arr = np.identity(
                1 + int(test_observed_arr.max() + (test_observed_arr.min() * -1))
            )[
                (test_observed_arr.reshape(test_observed_arr.shape[0], -1) + (test_observed_arr.min() * -1)).astype(int)
            ]

            for row in range(self.seq_len, test_observed_arr.shape[0]):
                test_observed_arr = test_observed_arr[row-self.seq_len:row]

                test_arr = torch.from_numpy(test_observed_arr, ctx=self.__ctx)
                test_arr = torch.unsqueeze(
                    test_arr,
                    dim=0
                )
                if test_batch_arr is None:
                    test_batch_arr = test_arr
                else:
                    test_batch_arr = torch.cat(
                        (
                            test_batch_arr,
                            test_arr,
                        ),
                        dim=0
                    )
                if test_batch_arr.shape[0] == self.batch_size:
                    yield None, None, test_batch_arr, None
                    test_batch_arr = None

    def pre_normalize(self, arr):
        '''
        Normalize before observation.

        Args:
            arr:    Tensor.
        
        Returns:
            Tensor.
        '''
        if self.norm_mode == "min_max":
            if torch.max(arr) != torch.min(arr):
                n = 0.0
            else:
                n = 1e-08
            arr = (arr - torch.min(arr)) / (torch.max(arr) - torch.min(arr) + n)
        elif self.norm_mode == "z_score":
            std = torch.std(arr)
            if std == 0:
                std += 1e-08
            arr = (arr - torch.mean(arr)) / std

        arr = arr * self.scale
        return arr



# -*- coding: utf-8 -*-
from accelbrainbase.controllable_model import ControllableModel
from accelbrainbase._mxnet._exception.init_deferred_error import InitDeferredError
from accelbrainbase.observabledata._mxnet.function_approximator import FunctionApproximator
from accelbrainbase.samplabledata.policy_sampler import PolicySampler
from accelbrainbase.computable_loss import ComputableLoss

from mxnet.gluon.block import HybridBlock
from mxnet import gluon
from mxnet import autograd
import numpy as np
import mxnet as mx
import mxnet.ndarray as nd

from mxnet import MXNetError
from logging import getLogger
from abc import abstractmethod


class DQLController(HybridBlock, ControllableModel):
    '''
    Abstract base class to implement the Deep Q-Learning.

    The structure of Q-Learning is based on the Epsilon Greedy Q-Leanring algorithm,
    which is a typical off-policy algorithm.  In this paradigm, stochastic searching 
    and deterministic searching can coexist by hyperparameter `epsilon_greedy_rate` 
    that is probability that agent searches greedy. Greedy searching is deterministic 
    in the sensethat policy of agent follows the selection that maximizes the Q-Value.

    References:
        - https://code.accel-brain.com/Reinforcement-Learning/README.html#deep-q-network
        - Egorov, M. (2016). Multi-agent deep reinforcement learning.(URL: https://pdfs.semanticscholar.org/dd98/9d94613f439c05725bad958929357e365084.pdf)
        - Gupta, J. K., Egorov, M., & Kochenderfer, M. (2017, May). Cooperative multi-agent control using deep reinforcement learning. In International Conference on Autonomous Agents and Multiagent Systems (pp. 66-83). Springer, Cham.
        - Mnih, V., Kavukcuoglu, K., Silver, D., Graves, A., Antonoglou, I., Wierstra, D., & Riedmiller, M. (2013). Playing atari with deep reinforcement learning. arXiv preprint arXiv:1312.5602.

    '''
    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    def __init__(
        self,
        function_approximator,
        policy_sampler,
        computable_loss,
        optimizer_name="SGD",
        learning_rate=1e-05,
        learning_attenuate_rate=1.0,
        attenuate_epoch=50,
        hybridize_flag=True,
        scale=1.0,
        ctx=mx.gpu(),
        initializer=None,
        recursive_learning_flag=False,
        **kwargs
    ):
        '''
        Init.

        Args:
            function_approximator:          is-a `FunctionApproximator`.
            policy_sampler:                 is-a `PolicySampler`.
            computable_loss:                is-a `ComputableLoss` or `mxnet.gluon.loss`.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
                                            
            optimizer_name:                 `str` of name of optimizer.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            scale:                          `float` of scaling factor for initial parameters.
            ctx:                            `mx.cpu()` or `mx.gpu()`.
            initializer:                    is-a `mxnet.initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.

        '''
        if isinstance(function_approximator, FunctionApproximator) is False:
            raise TypeError("The type of `function_approximator` must be `FunctionApproximator`.")
        if isinstance(policy_sampler, PolicySampler) is False:
            raise TypeError("The type of `policy_sampler` must be `PolicySampler`.")
        if isinstance(computable_loss, ComputableLoss) is False and isinstance(computable_loss, gluon.loss.Loss) is False:
            raise TypeError("The type of `computable_loss` must be `ComputableLoss` or `gluon.loss.Loss`.")

        super(DQLController, self).__init__()

        self.__function_approximator = function_approximator
        self.__policy_sampler = policy_sampler
        self.__computable_loss = computable_loss

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        if initializer is None:
            self.initializer = mx.initializer.Xavier(
                rnd_type="gaussian", 
                factor_type="in", 
                magnitude=1
            )
        else:
            if isinstance(initializer, mx.initializer.Initializer) is False:
                raise TypeError("The type of `initializer` must be `mxnet.initializer.Initializer`.")
            self.initializer = initializer

        if self.init_deferred_flag is False:
            try:
                self.collect_params().initialize(self.initializer, force_reinit=True, ctx=ctx)
                self.trainer = gluon.Trainer(
                    self.function_approximator.collect_params(), 
                    optimizer_name, 
                    {
                        "learning_rate": learning_rate
                    }
                )
                if hybridize_flag is True:
                    self.function_approximator.hybridize()

            except InitDeferredError:
                self.__logger.debug("The initialization should be deferred.")

        self.__learning_rate = learning_rate
        self.__learning_attenuate_rate = learning_attenuate_rate
        self.__attenuate_epoch = attenuate_epoch
        self.__recursive_learning_flag = recursive_learning_flag
        self.__q_logs_arr = np.array([])

        self.t = 0

    def collect_params(self, select=None):
        '''
        Overrided `collect_params` in `mxnet.gluon.HybridBlok`.
        '''
        params_dict = self.function_approximator.collect_params(select)
        return params_dict

    def learn(
        self, 
        iter_n=100,
    ):
        '''
        Learning.

        Args:
            iter_n:     `int` of the number of training iterations.
        '''
        q_logs_list = []
        r_logs_list = []
        posterior_logs_list = []

        learning_rate = self.__learning_rate
        state_arr = None
        state_meta_data_arr = None

        try:
            for n in range(iter_n):
                if (n + 1) % 100 == 0 or n < 100:
                    self.__logger.debug("-" * 100)
                    self.__logger.debug("Iter: " + str(n + 1))

                if ((n + 1) % self.__attenuate_epoch == 0):
                    learning_rate = learning_rate * self.__learning_attenuate_rate
                    self.trainer.set_learning_rate(learning_rate)

                # Draw samples of next possible actions from any distribution.
                # (batch, possible_n, dim1, dim2, ...)
                possible_action_arr, action_meta_data_arr = self.policy_sampler.draw()

                possible_reward_value_arr = None
                next_q_arr = None
                possible_predicted_q_arr = None

                for possible_i in range(possible_action_arr.shape[1]):
                    if action_meta_data_arr is not None:
                        meta_data_arr = action_meta_data_arr[:, possible_i]
                    else:
                        meta_data_arr = None

                    # Inference Q-Values.
                    _predicted_q_arr = self.function_approximator.inference(
                        possible_action_arr[:, possible_i]
                    )
                    if possible_predicted_q_arr is None:
                        possible_predicted_q_arr = nd.expand_dims(_predicted_q_arr, axis=1)
                    else:
                        possible_predicted_q_arr = nd.concat(
                            possible_predicted_q_arr,
                            nd.expand_dims(_predicted_q_arr, axis=1),
                            dim=1
                        )

                    # Observe reward values.
                    _reward_value_arr = self.policy_sampler.observe_reward_value(
                        state_arr, 
                        possible_action_arr[:, possible_i],
                        meta_data_arr=meta_data_arr,
                    )
                    if possible_reward_value_arr is None:
                        possible_reward_value_arr = nd.expand_dims(_reward_value_arr, axis=1)
                    else:
                        possible_reward_value_arr = nd.concat(
                            possible_reward_value_arr,
                            nd.expand_dims(_reward_value_arr, axis=1),
                            dim=1
                        )

                    # Inference the Max-Q-Value in next action time.
                    self.policy_sampler.observe_state(
                        state_arr=possible_action_arr[:, possible_i],
                        meta_data_arr=meta_data_arr
                    )
                    next_possible_action_arr, _ = self.policy_sampler.draw()
                    next_next_q_arr = None

                    for possible_j in range(next_possible_action_arr.shape[1]):
                        with autograd.predict_mode():
                            _next_next_q_arr = self.function_approximator.inference(
                                next_possible_action_arr[:, possible_j]
                            )
                        if next_next_q_arr is None:
                            next_next_q_arr = nd.expand_dims(
                                _next_next_q_arr,
                                axis=1
                            )
                        else:
                            next_next_q_arr = nd.concat(
                                next_next_q_arr,
                                nd.expand_dims(
                                    _next_next_q_arr, 
                                    axis=1
                                ),
                                dim=1
                            )

                    next_max_q_arr = next_next_q_arr.max(axis=1)

                    if next_q_arr is None:
                        next_q_arr = nd.expand_dims(
                            next_max_q_arr,
                            axis=1
                        )
                    else:
                        next_q_arr = nd.concat(
                            next_q_arr,
                            nd.expand_dims(
                                next_max_q_arr,
                                axis=1
                            ),
                            dim=1
                        )

                # Select action.
                selected_tuple = self.select_action(
                    possible_action_arr, 
                    possible_predicted_q_arr,
                    possible_reward_value_arr,
                    next_q_arr,
                    possible_meta_data_arr=action_meta_data_arr
                )
                action_arr, predicted_q_arr, reward_value_arr, next_q_arr, action_meta_data_arr = selected_tuple

                with autograd.record():
                    predicted_q_arr = self.function_approximator.inference(
                        action_arr
                    )
                    # Update real Q-Values.
                    real_q_arr = self.update_q(
                        reward_value_arr,
                        next_q_arr
                    )

                    if self.__q_logs_arr.shape[0] > 0:
                        self.__q_logs_arr = np.r_[
                            self.__q_logs_arr,
                            np.array([
                                predicted_q_arr.mean().asnumpy(), 
                                real_q_arr.mean().asnumpy()
                            ]).reshape(1, 2)
                        ]
                    else:
                        self.__q_logs_arr = np.array([
                            predicted_q_arr.mean().asnumpy(), 
                            real_q_arr.mean().asnumpy()
                        ]).reshape(1, 2)

                    # Learn Q-Values.
                    loss = self.computable_loss(
                        predicted_q_arr, 
                        real_q_arr
                    )
                if mx.nd.contrib.isnan(loss).astype(int).sum() == 0:
                    loss.backward()
                    self.trainer.step(predicted_q_arr.shape[0])
                    self.function_approximator.model.regularize()

                    if (n + 1) % 100 == 0 or n < 100:
                        self.__logger.debug("Reward value(mean): " + str(reward_value_arr.mean().asnumpy()[0]))
                        self.__logger.debug("Predicted Q-value(mean): " + str(predicted_q_arr.mean().asnumpy()[0]))
                        self.__logger.debug("Real Q-value(mean): " + str(real_q_arr.mean().asnumpy()[0]))
                        self.__logger.debug("Loss of Q-Value(mean): " + str(loss.mean().asnumpy()[0]))

                    if self.__recursive_learning_flag is True:
                        # Update State.
                        state_arr, state_meta_data_arr = self.policy_sampler.update_state(
                            action_arr, 
                            meta_data_arr=action_meta_data_arr
                        )
                        self.policy_sampler.observe_state(
                            state_arr=state_arr,
                            meta_data_arr=state_meta_data_arr
                        )
                else:
                    self.__logger.debug("The parameter update was skipped because the vanishing gradient problem or gradient explosion occurred.")

                # Epsode.
                self.t += 1

                if self.__recursive_learning_flag is True:
                    # Check.
                    end_flag = self.policy_sampler.check_the_end_flag(
                        state_arr, 
                        meta_data_arr=meta_data_arr
                    )
                    if end_flag is True:
                        break

        except KeyboardInterrupt:
            print("Keyboard Interrupt.")

    def inference(self, iter_n=100):
        '''
        Inference.

        Args:
            iter_n:     `int` of the number of training iterations.

        Returns:
            `list` of logs of states.
        '''
        state_arr_list = []
        q_value_arr_list = []
        state_arr = None
        try:
            for n in range(iter_n):
                # Draw samples of next possible actions from any distribution.
                # (batch, possible_n, dim1, dim2, ...)
                possible_action_arr, action_meta_data_arr = self.policy_sampler.draw()

                possible_reward_value_arr = None
                next_q_arr = None
                possible_predicted_q_arr = None

                for possible_i in range(possible_action_arr.shape[1]):
                    if action_meta_data_arr is not None:
                        meta_data_arr = action_meta_data_arr[:, possible_i]
                    else:
                        meta_data_arr = None

                    # Inference Q-Values.
                    _predicted_q_arr = self.function_approximator.inference(
                        possible_action_arr[:, possible_i]
                    )
                    if possible_predicted_q_arr is None:
                        possible_predicted_q_arr = nd.expand_dims(_predicted_q_arr, axis=1)
                    else:
                        possible_predicted_q_arr = nd.concat(
                            possible_predicted_q_arr,
                            nd.expand_dims(_predicted_q_arr, axis=1),
                            dim=1
                        )

                    # Observe reward values.
                    _reward_value_arr = self.policy_sampler.observe_reward_value(
                        state_arr, 
                        possible_action_arr[:, possible_i],
                        meta_data_arr=meta_data_arr,
                    )
                    if possible_reward_value_arr is None:
                        possible_reward_value_arr = nd.expand_dims(_reward_value_arr, axis=1)
                    else:
                        possible_reward_value_arr = nd.concat(
                            possible_reward_value_arr,
                            nd.expand_dims(_reward_value_arr, axis=1),
                            dim=1
                        )

                    # Inference the Max-Q-Value in next action time.
                    self.policy_sampler.observe_state(
                        state_arr=possible_action_arr[:, possible_i],
                        meta_data_arr=meta_data_arr
                    )
                    next_possible_action_arr, _ = self.policy_sampler.draw()
                    next_next_q_arr = None

                    for possible_j in range(next_possible_action_arr.shape[1]):
                        _next_next_q_arr = self.function_approximator.inference(
                            next_possible_action_arr[:, possible_j]
                        )
                        if next_next_q_arr is None:
                            next_next_q_arr = nd.expand_dims(
                                _next_next_q_arr,
                                axis=1
                            )
                        else:
                            next_next_q_arr = nd.concat(
                                next_next_q_arr,
                                nd.expand_dims(
                                    _next_next_q_arr, 
                                    axis=1
                                ),
                                dim=1
                            )

                    next_max_q_arr = next_next_q_arr.max(axis=1)

                    if next_q_arr is None:
                        next_q_arr = nd.expand_dims(
                            next_max_q_arr,
                            axis=1
                        )
                    else:
                        next_q_arr = nd.concat(
                            next_q_arr,
                            nd.expand_dims(
                                next_max_q_arr,
                                axis=1
                            ),
                            dim=1
                        )

                # Select action.
                selected_tuple = self.select_action(
                    possible_action_arr, 
                    possible_predicted_q_arr,
                    possible_reward_value_arr,
                    next_q_arr,
                    possible_meta_data_arr=action_meta_data_arr
                )
                action_arr, predicted_q_arr, reward_value_arr, next_q_arr, action_meta_data_arr = selected_tuple

                # Update State.
                state_arr, state_meta_data_arr = self.policy_sampler.update_state(
                    action_arr, 
                    meta_data_arr=action_meta_data_arr
                )
                self.policy_sampler.observe_state(
                    state_arr=state_arr,
                    meta_data_arr=state_meta_data_arr
                )
                state_arr_list.append(state_arr)
                q_value_arr_list.append(predicted_q_arr)

                # Epsode.
                self.t += 1

                # Check.
                end_flag = self.policy_sampler.check_the_end_flag(
                    state_arr, 
                    meta_data_arr=meta_data_arr
                )
                if end_flag is True:
                    break

        except KeyboardInterrupt:
            print("Keyboard Interrupt.")

        return state_arr_list, q_value_arr_list

    def extract_learned_dict(self):
        '''
        Extract (pre-) learned parameters.

        Returns:
            `dict` of the parameters.
        '''
        params_dict = self.collect_params()
        
        params_arr_dict = {}
        for k in params_dict:
            params_arr_dict.setdefault(k, params_dict[k].data())

        return params_arr_dict

    def save_parameters(self, filename):
        '''
        Save parameters to files.

        Args:
            filename:       File name.
        '''
        self.function_approximator.save_parameters(filename)

    def load_parameters(self, filename, ctx=None, allow_missing=False, ignore_extra=False):
        '''
        Load parameters to files.

        Args:
            filename:       File name.
            ctx:            `mx.cpu()` or `mx.gpu()`.
            allow_missing:  `bool` of whether to silently skip loading parameters not represents in the file.
            ignore_extra:   `bool` of whether to silently ignre parameters from the file that are not present in this `Block`.
        '''
        self.function_approximator.load_parameters(filename, ctx=ctx, allow_missing=allow_missing, ignore_extra=ignore_extra)

    @abstractmethod
    def select_action(
        self, 
        possible_action_arr, 
        possible_predicted_q_arr, 
        possible_reward_value_arr,
        possible_meta_data_arr=None
    ):
        '''
        Select action by Q(state, action).

        Args:
            possible_action_arr:        Tensor of actions.
            possible_predicted_q_arr:             Tensor of Q-Values.
            possible_reward_value_arr:       Tensor of reward values.
            possible_meta_data_arr:          Meta data of the actions.

        Retruns:
            Tuple(`np.ndarray` of action., Q-Value)
        '''
        raise NotImplementedError("This method must be implemented.")

    def update_q(self, reward_value_arr, next_max_q_arr):
        '''
        Update Q.
        
        Args:
            reward_value_arr:   `np.ndarray` of reward values.
            next_max_q_arr:     `np.ndarray` of maximum Q-Values in next time step.
        
        Returns:
            `np.ndarray` of real Q-Values.
        '''
        # Update Q-Value.
        return self.alpha_value * (reward_value_arr + (self.gamma_value * next_max_q_arr))

    __alpha_value = 1.0

    def get_alpha_value(self):
        '''
        getter
        Learning rate.
        '''
        if isinstance(self.__alpha_value, float) is False:
            raise TypeError("The type of __alpha_value must be float.")
        return self.__alpha_value

    def set_alpha_value(self, value):
        '''
        setter
        Learning rate.
        '''
        if isinstance(value, float) is False:
            raise TypeError("The type of __alpha_value must be float.")
        self.__alpha_value = value

    alpha_value = property(get_alpha_value, set_alpha_value)

    # Gamma Value.
    __gamma_value = 0.5

    def get_gamma_value(self):
        '''
        getter
        Gamma value.
        '''
        if isinstance(self.__gamma_value, float) is False:
            raise TypeError("The type of __gamma_value must be float.")
        return self.__gamma_value

    def set_gamma_value(self, value):
        '''
        setter
        Gamma value.
        '''
        if isinstance(value, float) is False:
            raise TypeError("The type of __gamma_value must be float.")
        self.__gamma_value = value

    gamma_value = property(get_gamma_value, set_gamma_value)

    __q_logs_arr = None

    def get_q_logs_arr(self):
        ''' getter '''
        return self.__q_logs_arr
    
    def set_q_logs_arr(self, values):
        ''' setter '''
        raise TypeError("The `q_logs_arr` must be read-only.")
    
    q_logs_arr = property(get_q_logs_arr, set_q_logs_arr)

    # is-a `PolicySampler`.
    __policy_sampler = None

    def get_policy_sampler(self):
        ''' getter for `PolicySampler` '''
        return self.__policy_sampler

    def set_policy_sampler(self, value):
        ''' setter for `PolicySampler` '''
        if isinstance(value, PolicySampler) is False:
            raise TypeError()
        self.__policy_sampler = value

    policy_sampler = property(get_policy_sampler, set_policy_sampler)

    # is-a `FunctionApproximator`.
    __function_approximator = None

    def get_function_approximator(self):
        ''' getter for `FunctionApproximator` '''
        return self.__function_approximator
    
    def set_function_approximator(self, value):
        ''' setter for `FunctionApproximator` '''
        if isinstance(value, FunctionApproximator) is False:
            raise TypeError()
        self.__function_approximator = value

    function_approximator = property(get_function_approximator, set_function_approximator)

    def get_computable_loss(self):
        ''' getter for `ComputableLoss`. '''
        return self.__computable_loss
    
    def set_computable_loss(self, value):
        ''' setter for `ComputableLoss`. '''
        if isinstance(value, ComputableLoss) is False and isinstance(value, gluon.loss.Loss) is False:
            raise TypeError("The type of `computable_loss` must be `ComputableLoss` or `gluon.loss.Loss`.")
        self.__computable_loss = value

    computable_loss = property(get_computable_loss, set_computable_loss)

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not. '''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not. '''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)



# -*- coding: utf-8 -*-
from accelbrainbase.iteratable_data import IteratableData
from accelbrainbase.samplabledata.pretext_sampler import PretextSampler
from abc import abstractmethod


class SSDATransformerIterator(IteratableData):
    '''
    Iterator for the Self-supervised Domain Adaptive Transformer.
    '''

    # `list` of `PretextSampler`s.
    __pretext_samplers_list = []

    def get_pretext_samplers_list(self):
        ''' getter for `list` of `PretextSampler`s. '''
        if isinstance(self.__pretext_samplers_list, list) is False:
            raise TypeError("The type of this property must be `list`.")
        if len(self.__pretext_samplers_list) == 0:
            raise ValueError("Number of elements in this property must be more than zero.")

        for i in range(len(self.__pretext_samplers_list)):
            if isinstance(self.__pretext_samplers_list[i], PretextSampler) is False:
                raise TypeError("The type of value of this property must be `PretextSampler`.")
        return self.__pretext_samplers_list

    def set_pretext_samplers_list(self, value):
        ''' setter for `list` of `PretextSampler`s. '''
        if isinstance(value, list) is False:
            raise TypeError("The type of this property must be `list`.")
        for i in range(len(value)):
            if isinstance(value[i], PretextSampler) is False:
                raise TypeError("The type of value of this property must be `PretextSampler`.")

        self.__pretext_samplers_list = value

    pretext_samplers_list = property(get_pretext_samplers_list, set_pretext_samplers_list)

    @abstractmethod
    def generate_learned_samples(self):
        '''
        Draw and generate learned samples.

        Returns:
            `Tuple` data. The shape is ...
            - encoder's observed data points in training.
            - decoder's observed data points in training.
            - encoder's masked data points in training.
            - decoder's masked data points in training.
            - encoder's observed data points in test.
            - decoder's observed data points in test.
            - encoder's masked data points in test.
            - decoder's masked data points in test.
            - Labeled data for downstream task in training.
            - Labeled data for downstream task in test.
            - `list` of encoder's observed data points in pretext task.
            - `list` of decoder's observed data points in pretext task.
            - `list` of encoder's masked data points in pretext task.
            - `list` of decoder's masked data points in pretext task.
            - `list` of labeled data in pretext task.
        '''
        raise NotImplementedError()

    def create_pretext_task_samples_list(self, target_domain_batch_arr):
        '''
        Create samples for pretext_task.

        Args:
            target_domain_batch_arr:    `nd.ndarray` of samples in target domain.
        
        Returns:
            Tuple data. The shape is ...
            - `list` of pretext task samples.
            - `list` of pretext task labels. The order of the elements corresponds to the pretext task samples.
        '''
        pretext_encoded_observed_arr_list = []
        pretext_decoded_observed_arr_list = []
        pretext_encoded_mask_arr_list = []
        pretext_decoded_mask_arr_list = []
        pretext_label_arr_list = []
        
        for i in range(len(self.pretext_samplers_list)):
            self.pretext_samplers_list[i].preprocess(target_domain_batch_arr)
            arr_tuple = self.pretext_samplers_list[i].draw()
            (
                pretext_encoded_observed_arr, 
                pretext_decoded_observed_arr, 
                pretext_encoded_mask_arr, 
                pretext_decoded_mask_arr, 
                pretext_label_arr
            ) = arr_tuple
            pretext_encoded_observed_arr_list.append(
                pretext_encoded_observed_arr
            )
            pretext_decoded_observed_arr_list.append(
                pretext_decoded_observed_arr
            )
            pretext_encoded_mask_arr_list.append(
                pretext_encoded_mask_arr
            )
            pretext_decoded_mask_arr_list.append(
                pretext_decoded_mask_arr
            )
            pretext_label_arr_list.append(
                pretext_label_arr
            )

        return (
            pretext_encoded_observed_arr_list,
            pretext_decoded_observed_arr_list,
            pretext_encoded_mask_arr_list,
            pretext_decoded_mask_arr_list,
            pretext_label_arr_list
        )

    def pre_normalize(self, arr):
        '''
        Normalize before observation.

        Args:
            arr:    Tensor.
        
        Returns:
            Tensor.
        '''
        if self.__norm_mode == "min_max":
            if arr.max() != arr.min():
                n = 0.0
            else:
                n = 1e-08
            arr = (arr - arr.min()) / (arr.max() - arr.min() + n)
        elif self.__norm_mode == "z_score":
            std = arr.asnumpy().std()
            if std == 0:
                std += 1e-08
            arr = (arr - arr.mean()) / std

        arr = arr * self.__scale
        return arr

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_epochs(self):
        ''' getter '''
        return self.__epochs

    def set_epochs(self, value):
        ''' setter '''
        self.__epochs = value

    epochs = property(get_epochs, set_epochs)

    def get_batch_size(self):
        ''' getter '''
        return self.__batch_size

    def set_batch_size(self, value):
        ''' setter '''
        self.__batch_size = value

    batch_size = property(get_batch_size, set_batch_size)

    __norm_mode = "z_score"

    def get_norm_mode(self):
        ''' getter '''
        return self.__norm_mode
    
    def set_norm_mode(self, value):
        ''' setter '''
        self.__norm_mode = value
    
    norm_mode = property(get_norm_mode, set_norm_mode)

    __scale = 1.0

    def get_scale(self):
        ''' getter '''
        return self.__scale
    
    def set_scale(self, value):
        ''' setter '''
        self.__scale = value
    
    scale = property(get_scale, set_scale)



# -*- coding: utf-8 -*-
from accelbrainbase.samplabledata.pretextsampler._torch.masked_language_sampler import MaskedLanguageSampler
import numpy as np
import torch


class MaskedSeq2SeqSampler(MaskedLanguageSampler):
    '''
    Sampler for pretext task based on the 
    Masked Sequence to Sequence Pre-training for Language Generation.

    References:
        - Song, K., Tan, X., Qin, T., Lu, J., & Liu, T. Y. (2019). Mass: Masked sequence to sequence pre-training for language generation. arXiv preprint arXiv:1905.02450.
    '''

    def __init__(
        self,
        masked_symbol=-1000,
        masked_seq_len=5,
    ):
        '''
        Init.

        Args:
            masked_symbol:      Symbol of mask.
            masked_seq_len:     The length of mask.
        '''
        self.masked_symbol = masked_symbol
        self.__masked_seq_len = masked_seq_len

    def preprocess(self, target_domain_arr):
        '''
        Preprocess observed data points in target domain.

        Args:
            target_domain_arr:      Tensor of observed data points in target domain.
                                    The rank of tensor is more than 2.
                                    The shape is (batch, sequence, token vector, ...).
        '''
        if self.__masked_seq_len == 1:
            super().preprocess(target_domain_arr)
            return

        pretext_label_arr = None
        target_domain_arr = target_domain_arr.float()
        pretext_encoded_observed_arr = target_domain_arr.detach()
        pretext_decoded_observed_arr = target_domain_arr.detach()
        for batch in range(target_domain_arr.shape[0]):
            target_seq = np.random.randint(
                low=0,
                high=target_domain_arr[batch].shape[0] - self.__masked_seq_len
            )
            if pretext_label_arr is None:
                pretext_label_arr = torch.unsqueeze(
                    target_domain_arr[batch, target_seq:target_seq+self.__masked_seq_len],
                    axis=0
                )
            else:
                pretext_label_arr = torch.cat(
                    (
                        pretext_label_arr,
                        torch.unsqueeze(
                            target_domain_arr[batch, target_seq:target_seq+self.__masked_seq_len],
                            axis=0
                        ),
                    ),
                    dim=0
                )
            target_domain_arr[batch, target_seq:target_seq+self.__masked_seq_len] = self.masked_symbol

        pretext_decoded_observed_arr = pretext_decoded_observed_arr * (target_domain_arr != self.masked_symbol).to(torch.int32) * self.masked_symbol

        self.pretext_encoded_observed_arr = target_domain_arr.float()
        self.pretext_decoded_observed_arr = pretext_decoded_observed_arr.float()
        self.pretext_encoded_mask_arr = None
        self.pretext_decoded_mask_arr = None
        self.pretext_label_arr = pretext_label_arr



# -*- coding: utf-8 -*-
from accelbrainbase.observabledata._torch.convolutional_neural_networks import ConvolutionalNeuralNetworks
import numpy as np
from logging import getLogger
import torch
from torch.optim.adam import Adam


class MobileNetV2(ConvolutionalNeuralNetworks):
    '''
    Mobilenet V2.

    References:
        - He, K., Zhang, X., Ren, S., & Sun, J. (2016). Deep residual learning for image recognition. In Proceedings of the IEEE conference on computer vision and pattern recognition (pp. 770-778).
        - Howard, A. G., Zhu, M., Chen, B., Kalenichenko, D., Wang, W., Weyand, T., ... & Adam, H. (2017). Mobilenets: Efficient convolutional neural networks for mobile vision applications. arXiv preprint arXiv:1704.04861.
        - Sandler, M., Howard, A., Zhu, M., Zhmoginov, A., & Chen, L. C. (2018). Mobilenetv2: Inverted residuals and linear bottlenecks. In Proceedings of the IEEE Conference on Computer Vision and Pattern Recognition (pp. 4510-4520).
    '''

    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    def __init__(
        self,
        computable_loss,
        optimizer_f=None,
        learning_rate=1e-05,
        weight_decay=0.01,
        ctx="cpu",
        activation="relu6",
        filter_multiplier=1.0,
        input_in_channel=3,
        input_filter_n=32,
        input_kernel_size=3,
        input_strides=2,
        input_padding=1,
        bottleneck_dict_list=[
            {
                "filter_rate": 1,
                "filter_n": 16,
                "block_n": 1,
                "stride": 1
            },
            {
                "filter_rate": 6,
                "filter_n": 24,
                "block_n": 2,
                "stride": 2
            },
            {
                "filter_rate": 6,
                "filter_n": 32,
                "block_n": 3,
                "stride": 2
            },
            {
                "filter_rate": 6,
                "filter_n": 64,
                "block_n": 4,
                "stride": 2
            },
            {
                "filter_rate": 6,
                "filter_n": 96,
                "block_n": 3,
                "stride": 1
            },
            {
                "filter_rate": 6,
                "filter_n": 160,
                "block_n": 3,
                "stride": 2
            },
            {
                "filter_rate": 6,
                "filter_n": 320,
                "block_n": 1,
                "stride": 1
            },
        ],
        hidden_filter_n=1280,
        pool_size=7,
        output_nn=None,
        shortcut_flag=True,
        global_shortcut_flag=False,
        output_batch_norm_flag=True,
        regularizatable_data_list=[],
        init_deferred_flag=None,
    ):
        '''
        Init.

        Args:
            computable_loss:            is-a `ComputableLoss` or `mxnet.gluon.loss`.
            initializer:                is-a `mxnet.initializer.Initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.

            ctx:                        `mx.cpu()` or `mx.gpu()`.
            hybridize_flag:             `bool` flag whether this class will hybridize models or not.
            activation:                 `str` of activtion function.
                                        - `relu`: ReLu function.
                                        - `relu6`: ReLU6 function.
                                        - `identity`: Identity function.
                                        - `identity_adjusted`: Identity function and normalization(divided by sum) function.

            filter_multiplier:          `float` of multiplier to compress size of model.
            input_filter_n:             `int` of the number of filters in input lauer.
            input_kernel:               `tuple` or `int` of kernel size in input layer.
            input_strides:              `tuple` or `int` of strides in input layer.
            input_padding:              `tuple` or `int` of zero-padding in input layer.
            bottleneck_dict_list:       `list` of information of bottleneck layers whose `dict` means ...
                                        - `filter_rate`: `float` of filter expfilter.
                                        - `filter_n`: `int` of the number of filters.
                                        - `block_n`: `int` of the number of blocks.
                                        - `stride`: `int` or `tuple` of strides.

            hidden_filter_n:            `int` of the number of filters in hidden layers.
            pool_size:                  `tuple` or `int` of pooling size in hidden layer.
                                        If `None`, the pooling layer will not attatched in hidden layer.

            optimizer_name:             `str` of name of optimizer.

            shortcut_flag:              `bool` flag that means shortcut will be added into residual blocks or not.
            global_shortcut_flag:       `bool` flag that means shortcut will be added into residual blocks or not.
                                        This shortcut will propagate input data into output layer.
            scale:                      `float` of scaling factor for initial parameters.
            init_deferred_flag:         `bool` that means initialization in this class will be deferred or not.
            regularizatable_data_list:  `list` of `RegularizatableData`.

        '''
        if init_deferred_flag is None:
            init_deferred_flag = self.init_deferred_flag
        elif isinstance(init_deferred_flag, bool) is False:
            raise TypeError("The type of `init_deferred_flag` must be `bool`.")

        self.init_deferred_flag = True

        super().__init__(
            computable_loss=computable_loss,
            learning_rate=learning_rate,
            output_nn=output_nn,
            ctx=ctx,
            not_init_flag=True,
            regularizatable_data_list=regularizatable_data_list,
            hidden_units_list=[],
            hidden_dropout_rate_list=[],
            hidden_batch_norm_list=[],
            hidden_activation_list=[],
        )
        self.init_deferred_flag = init_deferred_flag

        input_filter_n = int(round(input_filter_n * filter_multiplier))
        hidden_filter_n = int(hidden_filter_n * filter_multiplier) if int(filter_multiplier) > 1 else hidden_filter_n

        self.__input_layers_list = [
            torch.nn.Conv2d(
                in_channels=input_in_channel,
                out_channels=input_filter_n,
                kernel_size=input_kernel_size,
                stride=input_strides,
                padding=input_padding,
                bias=False,
            ),
            torch.nn.BatchNorm2d(
                input_filter_n,
            ),
        ]
        if activation == "relu6":
            self.__input_layers_list.append(
                torch.nn.ReLU6()
            )
        elif activation == "relu":
            self.__input_layers_list.append(
                torch.nn.ReLu()
            )
        self.__input_layers_list = torch.nn.ModuleList(self.__input_layers_list)

        inverted_residual_blocks_list = [None] * len(bottleneck_dict_list)
        in_filter_n = input_filter_n

        for i in range(len(bottleneck_dict_list)):
            inverted_residual_blocks_list[i] = []
            blocks_list = []

            filter_expfilter_n = int(round(in_filter_n * bottleneck_dict_list[i]["filter_rate"]))

            for j in range(bottleneck_dict_list[i]["block_n"]):
                if j == 0:
                    in_channels = input_filter_n
                else:
                    in_channels = filter_expfilter_n

                channel_expand_conv = torch.nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=filter_expfilter_n,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    groups=1,
                    bias=False,
                )
                if j == 0:
                    strides = bottleneck_dict_list[i]["stride"]
                else:
                    strides = 1

                bottleneck_conv = torch.nn.Conv2d(
                    in_channels=filter_expfilter_n,
                    out_channels=filter_expfilter_n,
                    kernel_size=3,
                    stride=strides,
                    padding=1,
                    groups=1,
                    bias=False,
                )
                linear_conv = torch.nn.Conv2d(
                    in_channels=filter_expfilter_n,
                    out_channels=bottleneck_dict_list[i]["filter_n"],
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    groups=1,
                    bias=False,
                )
                blocks_list.append(channel_expand_conv)
                blocks_list.append(
                    torch.nn.BatchNorm2d(
                        bottleneck_dict_list[i]["filter_n"]
                    ),
                )
                if activation == "relu6":
                    blocks_list.append(torch.nn.ReLU6())
                elif activation == "relu":
                    blocks_list.append(torch.nn.ReLu())

                blocks_list.append(bottleneck_conv)
                blocks_list.append(
                    torch.nn.BatchNorm2d(
                        bottleneck_dict_list[i]["filter_n"]
                    ),
                )
                if activation == "relu6":
                    blocks_list.append(torch.nn.ReLU6())
                elif activation == "relu":
                    blocks_list.append(torch.nn.ReLu())

                blocks_list.append(linear_conv)
                blocks_list.append(
                    torch.nn.BatchNorm2d(bottleneck_dict_list[i]["filter_n"]),
                )
                blocks_list = torch.nn.ModuleList(blocks_list)

            inverted_residual_blocks_list[i].append(blocks_list)
            inverted_residual_blocks_list[i] = torch.nn.ModuleList(inverted_residual_blocks_list[i])

            in_filter_n = int(round(bottleneck_dict_list[i]["filter_n"] * filter_multiplier))

        # TODO:ModuleListの2d 版
        self.__inverted_residual_blocks_list = torch.nn.ModuleList(inverted_residual_blocks_list)

        self.__output_layers_list = [
            torch.nn.Conv2d(
                in_channels=bottleneck_dict_list[i]["filter_n"],
                out_channels=hidden_filter_n,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            )
        ]

        if output_batch_norm_flag is True:
            self.__output_layers_list.append(
                torch.nn.BatchNorm2d(
                    hidden_filter_n
                )
            )

        if activation == "relu6":
            self.__output_layers_list.append(torch.nn.ReLU6())
        elif activation == "relu":
            self.__output_layers_list.append(torch.nn.ReLu())

        if pool_size is not None:
            self.__output_layers_list.append(torch.nn.AdaptiveAvgPool2d(pool_size))

        self.__output_layers_list = torch.nn.ModuleList(self.__output_layers_list)

        self.__shortcut_flag = shortcut_flag

        if output_nn is not None:
            self.output_nn = output_nn

        self.__global_shortcut_flag = global_shortcut_flag

        self.__ctx = ctx
        self.to(self.__ctx)

        if self.init_deferred_flag is False:
            if optimizer_f is not None:
                self.optimizer = optimizer_f(
                    self.parameters()
                )
            else:
                self.optimizer = Adam(
                    self.parameters(),
                    lr=learning_rate,
                    weight_decay=weight_decay,
                )

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        self.__activation = activation
        self.epoch = 0

    def inference(self, observed_arr):
        '''
        Inference the labels.

        Args:
            observed_arr:   rank-4 Array like or sparse matrix as the observed data points.
                            The shape is: (batch size, channel, height, width)

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        return self.forward(observed_arr)

    def forward(self, x):
        '''
        Forward with Gluon API.

        Args:
            F:      `mxnet.ndarray` or `mxnet.symbol`.
            x:      `mxnet.ndarray` of observed data points.
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        if self.__global_shortcut_flag is True:
            global_x = x

        for i in range(len(self.__input_layers_list)):
            x = self.__input_layers_list[i](x)

        if self.__activation == "identity_adjusted":
            x = x / torch.sum(torch.ones_like(x))

        for i in range(len(self.__inverted_residual_blocks_list)):
            for j in range(len(self.__inverted_residual_blocks_list[i])):
                if j >= 0 and self.__shortcut_flag is True:
                    _x = x
                for k in range(len(self.__inverted_residual_blocks_list[i][j])):
                    x = self.__inverted_residual_blocks_list[i][j][k](x)
                if j >= 0 and self.__shortcut_flag is True:
                    x = torch.add(_x, x)

        if self.__activation == "identity_adjusted":
            x = x / torch.sum(torch.ones_like(x))

        for i in range(len(self.__output_layers_list)):
            x = self.__output_layers_list[i](x)

        if self.__global_shortcut_flag is True:
            x = torch.add(global_x, x)

        self.feature_points_arr = x

        if self.output_nn is not None:
            x = self.output_nn(x)

        return x



# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod


class ExtractableData(metaclass=ABCMeta):
    '''
    The interface to extract samples from dataset.
    '''

    @abstractmethod
    def extract(
        self,
        path,
    ):
        '''
        Extract data.

        Args:
            path:       `str` of source path.
        
        Returns:
            Tensor.
        '''
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from accelbrainbase.regularizatabledata.constrain_weights import ConstrainWeights as _ConstrainWeights
import mxnet.ndarray as nd


class ConstrainWeights(_ConstrainWeights):
    '''
    Regularization for weights matrix
    to repeat multiplying the weights matrix and `0.9`
    until $\sum_{j=0}^{n}w_{ji}^2 < weight\_limit$.

    References:
        - Srivastava, N., Hinton, G., Krizhevsky, A., Sutskever, I., & Salakhutdinov, R. (2014). Dropout: a simple way to prevent neural networks from overfitting. The Journal of Machine Learning Research, 15(1), 1929-1958.
        - Zaremba, W., Sutskever, I., & Vinyals, O. (2014). Recurrent neural network regularization. arXiv preprint arXiv:1409.2329.
    '''

    def constrain_weight(self, weight_arr):
        square_weight_arr = weight_arr * weight_arr
        while nd.nansum(square_weight_arr) > self.weight_limit:
            weight_arr = weight_arr * 0.9
            square_weight_arr = weight_arr * weight_arr

        arr = weight_arr.asnumpy()
        return weight_arr



# -*- coding: utf-8 -*-
from accelbrainbase.controllable_model import ControllableModel
from accelbrainbase.observabledata._torch.transformer_model import TransformerModel
from accelbrainbase.observabledata._torch.transformermodel.transformer_encoder import TransformerEncoder
from accelbrainbase.observabledata._torch.transformermodel.transformer_decoder import TransformerDecoder
from accelbrainbase.observabledata._torch.transformermodel.transformer_reconstructor import TransformerReconstructor

from accelbrainbase.iteratabledata.transformer_iterator import TransformerIterator
from accelbrainbase.computable_loss import ComputableLoss
from accelbrainbase.regularizatable_data import RegularizatableData
import numpy as np
from logging import getLogger
import torch
from torch import nn
from torch.optim.adamw import AdamW


class TransformingAutoEncoderController(nn.Module, ControllableModel):
    '''
    Transforming Auto-Encoder.

    References:
        - Bahdanau, D., Cho, K., & Bengio, Y. (2014). Neural machine translation by jointly learning to align and translate. arXiv preprint arXiv:1409.0473.
        - Floridi, L., & Chiriatti, M. (2020). GPT-3: Its nature, scope, limits, and consequences. Minds and Machines, 30(4), 681-694.
        - Miller, A., Fisch, A., Dodge, J., Karimi, A. H., Bordes, A., & Weston, J. (2016). Key-value memory networks for directly reading documents. arXiv preprint arXiv:1606.03126.
        - Radford, A., Narasimhan, K., Salimans, T., & Sutskever, I. (2018) Improving Language Understanding by Generative Pre-Training. OpenAI (URL: https://s3-us-west-2.amazonaws.com/openai-assets/research-covers/language-unsupervised/language_understanding_paper.pdf)
        - Radford, A., Wu, J., Child, R., Luan, D., Amodei, D., & Sutskever, I. (2019). Language models are unsupervised multitask learners. OpenAI blog, 1(8), 9.
        - Vaswani, A., Shazeer, N., Parmar, N., Uszkoreit, J., Jones, L., Gomez, A. N., & Polosukhin, I. (2017). Attention is all you need. arXiv preprint arXiv:1706.03762.

    '''
    __loaded_filename = None
    __loaded_ctx = None

    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    def __init__(
        self,
        computable_loss=None,
        optimizer_f=None,
        encoder=None,
        decoder=None,
        reconstructor=None,
        layer_n=3,
        head_n=3,
        seq_len=5,
        depth_dim=100,
        hidden_dim=100,
        self_attention_activation_list=[],
        multi_head_attention_activation_list=[],
        fc_activation_list=[],
        learning_rate=1e-05,
        weight_decay=0.01,
        dropout_rate=0.5,
        ctx="cpu",
        regularizatable_data_list=[],
        not_init_flag=False,
    ):
        '''
        Init.

        Args:
            computable_loss:                is-a `ComputableLoss` or `gluon.loss`.
            encoder:                        is-a `TransformerModel`.
            decoder:                        is-a `TransformerModel`.
            reconstructor:                  is-a `TransformerModel`.
            layer_n:                        `int` of the number of layers.
            head_n:                         `int` of the number of heads for multi-head attention model.
            seq_len:                        `int` of the length of sequences.
            depth_dim:                      `int` of dimension of dense layer.
            hidden_dim:                     `int` of dimension of hidden(encoder) layer.
            self_attention_activation_list: `list` of `str` of activation function for self-attention model.
            multi_head_attention_activation_list:   `list` of `str` of activation function for multi-head attention model.
            fc_activation_list:             `list` of `str` of activation function in fully-connected layers.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
            optimizer_name:                 `str` of name of optimizer.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            scale:                          `float` of scaling factor for initial parameters.
            ctx:                            `mx.cpu()` or `mx.gpu()`.
            initializer:                    is-a `mxnet.initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.

        '''
        super(TransformingAutoEncoderController, self).__init__()

        if computable_loss is None:
            computable_loss = nn.CrossEntropyLoss()
        self.__computable_loss = computable_loss

        if encoder is None:
            if hidden_dim is None or hidden_dim == depth_dim:
                encoder = TransformerEncoder(
                    depth_dim=depth_dim,
                    layer_n=layer_n,
                    head_n=head_n,
                    self_attention_activation_list=self_attention_activation_list,
                    fc_activation_list=fc_activation_list,
                    computable_loss=computable_loss,
                    not_init_flag=not_init_flag,
                    dropout_rate=dropout_rate,
                    ctx=ctx
                )
            else:
                encoder = TransformerEncoder(
                    depth_dim=hidden_dim,
                    layer_n=layer_n,
                    head_n=head_n,
                    self_attention_activation_list=self_attention_activation_list,
                    fc_activation_list=fc_activation_list,
                    computable_loss=computable_loss,
                    not_init_flag=not_init_flag,
                    dropout_rate=dropout_rate,
                    ctx=ctx
                )
            encoder.embedding_flag = False
        else:
            if isinstance(encoder, TransformerModel) is False:
                raise TypeError("The type of `encoder` must be `TransformerModel`.")

        if decoder is None:
            if hidden_dim is None or hidden_dim == depth_dim:
                decoder = TransformerDecoder(
                    head_n=head_n,
                    depth_dim=depth_dim,
                    layer_n=layer_n,
                    self_attention_activation_list=self_attention_activation_list,
                    multi_head_attention_activation_list=multi_head_attention_activation_list,
                    fc_activation_list=fc_activation_list,
                    computable_loss=computable_loss,
                    not_init_flag=not_init_flag,
                    ctx=ctx
                )
            else:
                decoder = TransformerDecoder(
                    head_n=head_n,
                    depth_dim=hidden_dim,
                    output_dim=hidden_dim,
                    layer_n=layer_n,
                    self_attention_activation_list=self_attention_activation_list,
                    multi_head_attention_activation_list=multi_head_attention_activation_list,
                    fc_activation_list=fc_activation_list,
                    computable_loss=computable_loss,
                    not_init_flag=not_init_flag,
                    ctx=ctx
                )
            decoder.embedding_flag = False
        else:
            if isinstance(decoder, TransformerModel) is False:
                raise TypeError("The type of `decoder` must be `TransformerModel`.")

        if reconstructor is None:
            if hidden_dim is None or hidden_dim == depth_dim:
                reconstructor = TransformerReconstructor(
                    head_n=head_n,
                    depth_dim=depth_dim,
                    layer_n=layer_n,
                    self_attention_activation_list=self_attention_activation_list,
                    fc_activation_list=fc_activation_list,
                    computable_loss=computable_loss,
                    not_init_flag=not_init_flag,
                    ctx=ctx,
                )
            else:
                reconstructor = TransformerReconstructor(
                    head_n=head_n,
                    depth_dim=hidden_dim,
                    output_dim=depth_dim,
                    layer_n=layer_n,
                    self_attention_activation_list=self_attention_activation_list,
                    fc_activation_list=fc_activation_list,
                    computable_loss=computable_loss,
                    not_init_flag=not_init_flag,
                    ctx=ctx,
                )
            reconstructor.embedding_flag = False
        else:
            if isinstance(reconstructor, TransformerModel) is False:
                raise TypeError("The type of `reconstructor` must be `TransformerModel`.")

        logger = getLogger("accelbrainbase")
        self.logger = logger

        self.encoder = encoder
        self.decoder = decoder
        self.reconstructor = reconstructor

        if hidden_dim is not None and hidden_dim != depth_dim:
            self.encoder_hidden_fc = nn.Linear(
                depth_dim,
                hidden_dim, 
                bias=True,
            )
            self.decoder_hidden_fc = nn.Linear(
                depth_dim,
                hidden_dim, 
                bias=True,
            )
            init_flag = True
        else:
            self.encoder_hidden_fc = None
            self.decoder_hidden_fc = None
            init_flag = False

        self.__ctx = ctx
        self.to(self.__ctx)

        if self.init_deferred_flag is False:
            if not_init_flag is False:
                if optimizer_f is not None:
                    if init_flag is True:
                        self.optimizer = optimizer_f(
                            self.parameters(),
                        )
                else:
                    if init_flag is True:
                        self.optimizer = AdamW(
                            self.parameters(),
                            lr=learning_rate,
                            weight_decay=weight_decay
                        )

        for v in regularizatable_data_list:
            if isinstance(v, RegularizatableData) is False:
                raise TypeError("The type of values of `regularizatable_data_list` must be `RegularizatableData`.")
        self.__regularizatable_data_list = regularizatable_data_list

        self.__learning_rate = learning_rate
        self.__weight_decay = weight_decay
        self.seq_len = seq_len
        self.epoch = 0

    def learn(self, iteratable_data):
        '''
        Learn samples drawn by `IteratableData.generate_learned_samples()`.

        Args:
            iteratable_data:     is-a `TransformerIterator`.
        '''
        if isinstance(iteratable_data, TransformerIterator) is False:
            raise TypeError("The type of `iteratable_data` must be `TransformerIterator`.")

        self.__loss_list = []
        learning_rate = self.__learning_rate

        try:
            epoch = self.epoch
            iter_n = 0
            for encoded_observed_arr, decoded_observed_arr, encoded_mask_arr, decoded_mask_arr, test_encoded_observed_arr, test_decoded_observed_arr, test_encoded_mask_arr, test_decoded_mask_arr, training_target_arr, test_target_arr in iteratable_data.generate_learned_samples():
                self.epoch = epoch
                if self.encoder.optimizer is not None and self.decoder.optimizer is not None:
                    optimizer_setup_flag = True
                    self.encoder.optimizer.zero_grad()
                    self.decoder.optimizer.zero_grad()
                    self.optimizer.zero_grad()
                else:
                    optimizer_setup_flag = False

                pred_arr = self.inference(
                    encoded_observed_arr, 
                    decoded_observed_arr, 
                    encoded_mask_arr, 
                    decoded_mask_arr
                )
                loss = self.compute_loss(
                    pred_arr,
                    training_target_arr
                )
                if optimizer_setup_flag is False:
                    self.encoder.optimizer.zero_grad()
                    self.decoder.optimizer.zero_grad()
                    self.optimizer.zero_grad()
                    pred_arr = self.inference(
                        encoded_observed_arr, 
                        decoded_observed_arr, 
                        encoded_mask_arr, 
                        decoded_mask_arr
                    )
                    loss = self.compute_loss(
                        pred_arr,
                        training_target_arr
                    )

                loss.backward()
                self.optimizer.step()
                self.decoder.optimizer.step()
                self.encoder.optimizer.step()
                self.regularize()
                self.decoder.regularize()
                self.encoder.regularize()

                if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                    if torch.inference_mode():
                        test_pred_arr = self.inference(
                            test_encoded_observed_arr, 
                            test_decoded_observed_arr, 
                            test_encoded_mask_arr, 
                            test_decoded_mask_arr
                        )

                        test_loss = self.compute_loss(
                            test_pred_arr,
                            test_target_arr
                        )

                    _loss = loss.to('cpu').detach().numpy().copy()
                    _test_loss = test_loss.to('cpu').detach().numpy().copy()

                    self.__loss_list.append((_loss, _test_loss))
                    self.logger.debug("Epochs: " + str(epoch + 1) + " Train loss: " + str(_loss) + " Test loss: " + str(_test_loss))
                    epoch += 1
                iter_n += 1

        except KeyboardInterrupt:
            self.logger.debug("Interrupt.")

        self.logger.debug("end. ")
        self.epoch = epoch

    def inference(
        self, 
        encoded_observed_arr, 
        decoded_observed_arr, 
        encoded_mask_arr=None,
        decoded_mask_arr=None,
    ):
        '''
        Inference samples drawn by `IteratableData.generate_inferenced_samples()`.

        Args:
            encoded_observed_arr:   rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            decoded_observed_arr:   rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            encoded_mask_arr:       rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            decoded_mask_arr:       rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        return self(
            encoded_observed_arr, 
            decoded_observed_arr, 
            encoded_mask_arr,
            decoded_mask_arr,
        )

    def compute_loss(self, pred_arr, labeled_arr):
        '''
        Compute loss.

        Args:
            pred_arr:       `mxnet.ndarray` or `mxnet.symbol`.
            labeled_arr:    `mxnet.ndarray` or `mxnet.symbol`.

        Returns:
            loss.
        '''
        return self.__computable_loss(pred_arr, labeled_arr)

    def forward(
        self, 
        encoded_observed_arr, 
        decoded_observed_arr, 
        encoded_mask_arr=None,
        decoded_mask_arr=None,
    ):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:                      `mxnet.ndarray` or `mxnet.symbol`.
            encoded_observed_arr:   rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            decoded_observed_arr:   rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            encoded_mask_arr:       rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            decoded_mask_arr:       rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        if self.__loaded_filename is not None:
            loaded_filename = self.__loaded_filename
            self.__loaded_filename = None
            init_encoded_observed_arr = encoded_observed_arr.detach()
            init_decoded_observed_arr = decoded_observed_arr.detach()
            if encoded_mask_arr is not None:
                init_encoded_mask_arr = encoded_mask_arr.detach()
            else:
                init_encoded_mask_arr = None
            if decoded_mask_arr is not None:
                init_decoded_mask_arr = decoded_mask_arr.detach()
            else:
                init_decoded_mask_arr = decoded_mask_arr

            _ = self.forward(
                init_encoded_observed_arr, 
                init_decoded_observed_arr, 
                init_encoded_mask_arr,
                init_decoded_mask_arr,
            )
            self.load_parameters(loaded_filename, ctx=self.__loaded_ctx)
            self.__loaded_ctx = None

        if encoded_mask_arr is None:
            encoded_mask_arr = torch.ones(
                (encoded_observed_arr.shape[0], 1, 1, 1), 
            )
            encoded_mask_arr = encoded_mask_arr.to(encoded_observed_arr.device)
        if decoded_mask_arr is None:
            decoded_mask_arr = torch.ones(
                (decoded_observed_arr.shape[0], 1, 1, 1), 
            )
            decoded_mask_arr = decoded_mask_arr.to(decoded_observed_arr.device)

        if self.encoder_hidden_fc is not None:
            encoded_observed_arr = self.encoder_hidden_fc(encoded_observed_arr)
        if self.decoder_hidden_fc is not None:
            decoded_observed_arr = self.decoder_hidden_fc(decoded_observed_arr)

        encoded_arr = self.encoder(
            encoded_observed_arr,
            encoded_mask_arr
        )
        decoded_arr = self.decoder(
            decoded_observed_arr, 
            encoded_arr, 
            decoded_mask_arr,
            encoded_mask_arr,
        )

        self.feature_points_arr = decoded_arr

        reconstructed_arr = self.reconstructor(
            decoded_arr,
            None
        )

        return reconstructed_arr

    def extract_learned_dict(self):
        '''
        Extract (pre-) learned parameters.

        Returns:
            `dict` of the parameters.
        '''
        params_dict = {}
        for k in self.state_dict().keys():
            params_dict.setdefault(k, self.state_dict()[k])

        return params_dict

    def regularize(self):
        '''
        Regularization.
        '''
        if len(self.__regularizatable_data_list) > 0:
            params_dict = self.extract_learned_dict()
            for regularizatable in self.__regularizatable_data_list:
                params_dict = regularizatable.regularize(params_dict)

            for k, params in params_dict.items():
                self.load_state_dict({k: params}, strict=False)

    def __rename_file(self, filename):
        filename_list = filename.split(".")
        _format = filename_list[-1]
        encoder_filename = filename.replace("." + _format, "_encoder." + _format)
        decoder_filename = filename.replace("." + _format, "_decoder." + _format)
        reconstructor_filename = filename.replace("." + _format, "_reconstructor." + _format)
        return encoder_filename, decoder_filename, reconstructor_filename

    def save_parameters(self, filename):
        '''
        Save parameters to files.

        Args:
            filename:       File name.
        '''
        encoder_filename, decoder_filename, reconstructor_filename = self.__rename_file(filename)

        self.encoder.epoch = self.epoch
        self.encoder.loss_arr = self.loss_arr
        self.encoder.save_parameters(encoder_filename)

        self.decoder.epoch = self.epoch
        self.decoder.loss_arr = self.loss_arr
        self.decoder.save_parameters(decoder_filename)

        self.reconstructor.epoch = self.epoch
        self.reconstructor.loss_arr = self.loss_arr
        self.reconstructor.save_parameters(decoder_filename)

        torch.save(
            {
                'model_state_dict': self.state_dict(),
                'optimizer_state_dict': self.optimizer.state_dict(),
                'epoch': self.epoch,
                'loss': self.loss_arr,
            }, 
            filename
        )

    def load_parameters(self, filename, ctx=None, strict=True):
        '''
        Load parameters to files.

        Args:
            filename:       File name.
            ctx:            Context-manager that changes the selected device.
            strict:         Whether to strictly enforce that the keys in state_dict match the keys returned by this module’s state_dict() function. Default: `True`.
        '''
        try:
            encoder_filename, decoder_filename, reconstructor_filename = self.__rename_file(filename)
            self.encoder.load_parameters(encoder_filename, ctx=ctx, strict=strict)
            self.decoder.load_parameters(decoder_filename, ctx=ctx, strict=strict)
            self.reconstructor.load_parameters(reconstructor_filename, ctx=ctx, strict=strict)

            checkpoint = torch.load(filename)
            self.load_state_dict(checkpoint['model_state_dict'], strict=strict)
            self.optimizer.load_state_dict(
                checkpoint['optimizer_state_dict']
            )
            self.epoch = checkpoint['epoch']
            self.__loss_list = checkpoint['loss'].tolist()
        except RuntimeError:
            self.__loaded_filename = filename
            self.__loaded_ctx = ctx

        if ctx is not None:
            self.to(ctx)
            self.encoder.to(ctx)
            self.decoder.to(ctx)
            self.reconstructor.to(ctx)
            self.__ctx = ctx

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not.'''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not.'''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)

    __loss_list = []

    def get_loss_arr(self):
        ''' getter '''
        return np.array(self.__loss_list)
    
    def set_loss_arr(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")
    
    loss_arr = property(get_loss_arr, set_loss_arr)



# -*- coding: utf-8 -*-
from accelbrainbase.noiseable_data import NoiseableData
import numpy as np


class CutoutNoise(NoiseableData):
    '''
    Gauss noise function.

    References:
        - DeVries, T., & Taylor, G. W. (2017). Improved regularization of convolutional neural networks with cutout. arXiv preprint arXiv:1708.04552.
    '''

    def __init__(self, height=10, width=10):
        '''
        Init.

        Args:

        '''
        self.__height = height
        self.__width = width

    def noise(self, arr):
        '''
        Noise.

        Args:
            arr:    Tensor (4 or 5 rank).
        
        Returns:
            Tensor.
        '''
        section_arr = np.ones(arr.shape)
        if section_arr.ndim == 4:
            start_hight = np.random.randint(low=0, high=section_arr.shape[2] - self.__height)
            start_width = np.random.randint(low=0, high=section_arr.shape[3] - self.__width)
            section_arr[:, :, start_hight:start_hight+self.__height, start_width:start_width+self.__width] = 0
        elif section_arr.ndim == 5:
            start_hight = np.random.randint(low=0, high=section_arr.shape[3] - self.__height)
            start_width = np.random.randint(low=0, high=section_arr.shape[4] - self.__width)
            section_arr[:, :, :, start_hight:start_hight+self.__height, start_width:start_width+self.__width] = 0
        return arr * section_arr



# -*- coding: utf-8 -*-
from accelbrainbase.observabledata._torch.neural_networks import NeuralNetworks
from accelbrainbase.iteratable_data import IteratableData
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.optim.optimizer import Optimizer, required
from torch.optim.adam import Adam
from logging import getLogger


class AutoEncoder(NeuralNetworks):
    '''
    Auto-Encoder.

    References:
        - Kamyshanska, H., & Memisevic, R. (2014). The potential energy of an autoencoder. IEEE transactions on pattern analysis and machine intelligence, 37(6), 1261-1273.
    '''

    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    def __init__(
        self,
        encoder,
        decoder,
        computable_loss,
        learning_rate=1e-05,
        ctx="cpu",
        tied_weights_flag=False,
        not_init_flag=False,
    ):
        '''
        Init.

        Args:
            encoder:                        is-a `NeuralNetworks`.
            decoder:                        is-a `NeuralNetworks`.
            computable_loss:                is-a `ComputableLoss` or `mxnet.gluon.loss`.
            learning_rate:                  `float` of learning rate.
            observed_activation:            `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` 
                                            that activates observed data points.

            ctx:                            Context-manager that changes the selected device.
            tied_weights_flag:              `bool` of flag to tied weights or not.
            not_init_flag:                  `bool` of whether initialize parameters or not.
        '''
        if isinstance(encoder, NeuralNetworks) is False:
            raise TypeError("The type of `encoder` must be `NeuralNetworks`.")
        if isinstance(decoder, NeuralNetworks) is False:
            raise TypeError("The type of `decoder` must be `NeuralNetworks`.")

        init_deferred_flag = self.init_deferred_flag
        self.init_deferred_flag = True
        super(AutoEncoder, self).__init__(computable_loss=computable_loss)
        self.init_deferred_flag = init_deferred_flag

        logger = getLogger("accelbrainbase")
        self.__logger = logger
        self.encoder = encoder
        self.decoder = decoder
        self.__tied_weights_flag = tied_weights_flag

        self.__computable_loss = computable_loss

        self.encoder_optimizer = None
        self.decoder_optimizer = None

        self.epoch = 0
        self.__learning_rate = learning_rate
        self.__not_init_flag = not_init_flag

        self.__ctx = ctx
        self.__encoder_input_dim = None
        self.__decoder_input_dim = None
        self.__loss_list = []

    def parameters(self):
        '''
        '''
        return [
            {
                "params": self.encoder.parameters(),
            },
            {
                "params": self.decoder.parameters(),
            }
        ]

    def initialize_params(self, input_dim):
        '''
        Initialize params.
        '''
        if self.encoder_optimizer is not None and self.decoder_optimizer is not None:
            return

        self.__encoder_input_dim = input_dim
        self.__decoder_input_dim = self.encoder.units_list[-1]
        if self.init_deferred_flag is False:
            if self.__not_init_flag is False:
                self.encoder.initialize_params(self.__encoder_input_dim)
                self.decoder.initialize_params(self.__decoder_input_dim)
                self.encoder_optimizer = self.encoder.optimizer
                self.decoder_optimizer = self.decoder.optimizer

    def learn(self, iteratable_data):
        '''
        Learn samples drawn by `IteratableData.generate_learned_samples()`.

        Args:
            iteratable_data:     is-a `IteratableData`.
        '''
        if isinstance(iteratable_data, IteratableData) is False:
            raise TypeError("The type of `iteratable_data` must be `IteratableData`.")

        self.__loss_list = []
        try:
            epoch = self.epoch
            iter_n = 0
            for batch_observed_arr, batch_target_arr, test_batch_observed_arr, test_batch_target_arr in iteratable_data.generate_learned_samples():
                batch_size = batch_observed_arr.shape[0]
                self.initialize_params(
                    input_dim=batch_observed_arr.reshape(batch_size, -1).shape[1]
                )
                self.encoder_optimizer.zero_grad()
                self.decoder_optimizer.zero_grad()
                # rank-3
                pred_arr = self.inference(batch_observed_arr)
                loss = self.compute_loss(
                    pred_arr,
                    batch_target_arr
                )
                loss.backward()
                self.encoder_optimizer.step()
                self.decoder_optimizer.step()
                self.regularize()

                if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                    with torch.inference_mode():
                        # rank-3
                        test_pred_arr = self.inference(test_batch_observed_arr)

                        test_loss = self.compute_loss(
                            test_pred_arr,
                            test_batch_target_arr
                        )
                    _loss = loss.to('cpu').detach().numpy().copy()
                    _test_loss = test_loss.to('cpu').detach().numpy().copy()
                    self.__loss_list.append((_loss, _test_loss))
                    self.__logger.debug("Epochs: " + str(epoch + 1) + " Train loss: " + str(_loss) + " Test loss: " + str(_test_loss))
                    epoch += 1
                iter_n += 1

        except KeyboardInterrupt:
            self.__logger.debug("Interrupt.")

        self.epoch = epoch
        self.__logger.debug("end. ")

    def inference(self, observed_arr):
        '''
        Inference the feature points.

        Args:
            observed_arr:   rank-2 Array like or sparse matrix as the observed data points.
                            The shape is: (batch size, feature points)

        Returns:
            `tensor` of inferenced feature points.
        '''
        pred_arr = self.forward(observed_arr)
        return pred_arr

    def compute_loss(self, pred_arr, labeled_arr):
        '''
        Compute loss.

        Args:
            pred_arr:       `tensor`.
            labeled_arr:    `tensor`.

        Returns:
            loss.
        '''
        return self.__computable_loss(pred_arr, labeled_arr)

    def extract_feature_points(self):
        '''
        Extract the activities in hidden layer and reset it, 
        considering this method will be called per one cycle in instances of time-series.

        Returns:
            The `mxnet.ndarray` of array like or sparse matrix of feature points or virtual visible observed data points.
        '''
        return self.feature_points_arr

    def extract_learned_dict(self):
        '''
        Extract (pre-) learned parameters.

        Returns:
            `dict` of the parameters.
        '''
        params_arr_dict = {}

        params_dict = self.encoder.extract_learned_dict()
        for k in params_dict:
            params_arr_dict.setdefault(k, params_dict[k].data())

        params_dict = self.decoder.extract_learned_dict()
        for k in params_dict:
            params_arr_dict.setdefault(k, params_dict[k].data())

        return params_arr_dict

    def forward(self, x):
        '''
        Forward with torch.

        Args:
            x:      `tensor` of observed data points.
        
        Returns:
            `tensor` of inferenced feature points.
        '''
        batch_size = x.shape[0]
        self.initialize_params(
            input_dim=x.reshape(batch_size, -1).shape[1]
        )

        encoded_arr = self.encoder.inference(x)
        self.feature_points_arr = encoded_arr
        decoded_arr = self.decoder.inference(encoded_arr)
        self.__pred_arr = decoded_arr
        return decoded_arr

    def regularize(self):
        '''
        Regularization.
        '''
        self.encoder.regularize()
        self.decoder.regularize()
        self.__tie_weights()

    def __tie_weights(self):
        if self.__tied_weights_flag is True:
            encoder_params_dict = self.encoder.extract_learned_dict()
            decoder_params_dict = self.decoder.extract_learned_dict()
            encoder_weight_keys_list = [key for key in encoder_params_dict.keys() if "fc_list" in key and "weight" in key]
            decoder_weight_keys_list = [key for key in decoder_params_dict.keys() if "fc_list" in key and "weight" in key]

            if len(encoder_weight_keys_list) != len(decoder_weight_keys_list):
                raise ValueError(
                    "The number of layers is invalid."
                )

            for i in range(len(self.encoder.units_list)):
                encoder_layer = i
                decoder_layer = len(self.encoder.units_list) - i - 1
                encoder_weight_keys, decoder_weight_keys = None, None
                for _encoder_weight_keys in encoder_weight_keys_list:
                    if "fc_list." + str(encoder_layer) + ".weight" in _encoder_weight_keys:
                        encoder_weight_keys = _encoder_weight_keys
                        break

                for _decoder_weight_keys in decoder_weight_keys_list:
                    if "fc_list." + str(decoder_layer) + ".weight" in _decoder_weight_keys:
                        decoder_weight_keys = _decoder_weight_keys
                        break

                if encoder_weight_keys is not None and decoder_weight_keys is not None:
                    try:
                        decoder_params_dict[decoder_weight_keys] = encoder_params_dict[encoder_weight_keys].T
                    except AssertionError:
                        raise ValueError(
                            "The shapes of weight matrixs must be equivalents in encoder layer " + str(encoder_layer) + " and decoder layer " + str(decoder_layer)
                        )

            for k, params in decoder_params_dict.items():
                if k in decoder_weight_keys_list:
                    self.decoder.load_state_dict({k: params}, strict=False)

    def __rename_file(self, filename):
        filename_list = filename.split(".")
        _format = filename_list[-1]
        encoder_filename = filename.replace("." + _format, "_encoder." + _format)
        decoder_filename = filename.replace("." + _format, "_decoder." + _format)
        return encoder_filename, decoder_filename

    def save_parameters(self, filename):
        '''
        Save parameters to files.

        Args:
            filename:       File name.
        '''
        encoder_filename, decoder_filename = self.__rename_file(filename)
        self.encoder.save_parameters(encoder_filename)
        self.decoder.save_parameters(decoder_filename)
        torch.save(
            {
                'epoch': self.epoch,
                'loss': self.loss_arr,
            }, 
            filename
        )

    def load_parameters(self, filename, ctx=None, strict=True):
        '''
        Load parameters to files.

        Args:
            filename:       File name.
            ctx:            Context-manager that changes the selected device.
            strict:         Whether to strictly enforce that the keys in state_dict match the keys returned by this module’s state_dict() function. Default: `True`.
        '''
        encoder_filename, decoder_filename = self.__rename_file(filename)
        self.encoder.load_parameters(encoder_filename, ctx=ctx, strict=strict)
        self.decoder.load_parameters(decoder_filename, ctx=ctx, strict=strict)

        checkpoint = torch.load(filename)
        self.epoch = checkpoint['epoch']
        self.__loss_list = checkpoint['loss'].tolist()

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_loss_arr(self):
        ''' getter for losses. '''
        return np.array(self.__loss_list)

    loss_arr = property(get_loss_arr, set_readonly)

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not.'''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not.'''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)



# -*- coding: utf-8 -*-
from accelbrainbase.observable_data import ObservableData
from accelbrainbase.iteratable_data import IteratableData
from accelbrainbase.regularizatable_data import RegularizatableData
from accelbrainbase.computable_loss import ComputableLoss
import numpy as np
from logging import getLogger
import torch
from torch import nn
import torch.nn.functional as F
from torch.optim.optimizer import Optimizer, required
from torch.optim.adam import Adam


class LSTMNetworks(nn.Module, ObservableData):
    '''
    Long short term memory(LSTM) networks.
    
    Originally, Long Short-Term Memory(LSTM) networks as a 
    special RNN structure has proven stable and powerful for 
    modeling long-range dependencies.
    
    The Key point of structural expansion is its memory cell 
    which essentially acts as an accumulator of the state information. 
    Every time observed data points are given as new information and 
    input to LSTM's input gate, its information will be accumulated to 
    the cell if the input gate is activated. The past state of cell 
    could be forgotten in this process if LSTM's forget gate is on.
    Whether the latest cell output will be propagated to the final state 
    is further controlled by the output gate.
    
    References:
        - Cho, K., Van Merriënboer, B., Gulcehre, C., Bahdanau, D., Bougares, F., Schwenk, H., & Bengio, Y. (2014). Learning phrase representations using RNN encoder-decoder for statistical machine translation. arXiv preprint arXiv:1406.1078.
        - Malhotra, P., Ramakrishnan, A., Anand, G., Vig, L., Agarwal, P., & Shroff, G. (2016). LSTM-based encoder-decoder for multi-sensor anomaly detection. arXiv preprint arXiv:1607.00148.
        - Zaremba, W., Sutskever, I., & Vinyals, O. (2014). Recurrent neural network regularization. arXiv preprint arXiv:1409.2329.

    '''

    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    def __init__(
        self,
        computable_loss,
        initializer_f=None,
        optimizer_f=None,
        learning_rate=1e-05,
        seq_len=None,
        hidden_n=200,
        output_n=1,
        dropout_rate=0.5,
        input_adjusted_flag=True,
        observed_activation=torch.nn.Tanh(),
        input_gate_activation=torch.nn.Sigmoid(),
        forget_gate_activation=torch.nn.Sigmoid(),
        output_gate_activation=torch.nn.Sigmoid(),
        hidden_activation=torch.nn.Tanh(),
        output_activation=torch.nn.Tanh(),
        output_layer_flag=True,
        output_no_bias_flag=False,
        output_nn=None,
        ctx="cpu",
        regularizatable_data_list=[],
        not_init_flag=False,
    ):
        '''
        Init.

        Args:
            computable_loss:                is-a `ComputableLoss` or `mxnet.gluon.loss`.
            initializer:                    is-a `mxnet.initializer.Initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            batch_size:                     `int` of batch size of mini-batch.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
                                            

            seq_len:                        `int` of the length of sequences.
                                            This means refereed maxinum step `t` in feedforward.
                                            If `0`, this model will reference all series elements included 
                                            in observed data points.
                                            If not `0`, only first sequence will be observed by this model 
                                            and will be feedfowarded as feature points.
                                            This parameter enables you to build this class as `Decoder` in
                                            Sequence-to-Sequence(Seq2seq) scheme.

            hidden_n:                       `int` of the number of units in hidden layer.
            output_n:                       `int` of the nuber of units in output layer.
            dropout_rate:                   `float` of dropout rate.
            input_adjusted_flag:            `bool` of flag that means this class will adjusted observed data points by normalization.
            observed_activation:            `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` 
                                            that activates observed data points.

            optimizer_name:                 `str` of name of optimizer.

            input_gate_activation:          `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in input gate.
            forget_gate_activation:         `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in forget gate.
            output_gate_activation:         `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in output gate.
            hidden_activation:              `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in hidden layer.
            output_activation:              `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in output layer.
                                            If this value is `identity`, the activation function equivalents to the identity function.

            output_layer_flag:              `bool` that means this class has output layer or not.
            output_no_bias_flag:            `bool` for using bias or not in output layer(last hidden layer).
            output_nn:               is-a `NNHybrid` as output layers.
                                            If not `None`, `output_layer_flag` and `output_no_bias_flag` will be ignored.

            ctx:                            `mx.cpu()` or `mx.gpu()`.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            regularizatable_data_list:           `list` of `Regularizatable`.
            scale:                          `float` of scaling factor for initial parameters.
        '''
        if isinstance(computable_loss, ComputableLoss) is False and isinstance(computable_loss, nn.modules.loss._Loss) is False:
            raise TypeError("The type of `computable_loss` must be `ComputableLoss` or `gluon.loss.Loss`.")

        super(LSTMNetworks, self).__init__()
        self.initializer_f = initializer_f
        self.optimizer_f = optimizer_f
        self.__not_init_flag = not_init_flag

        if dropout_rate > 0.0:
            self.dropout_forward = nn.Dropout(p=dropout_rate)
        else:
            self.dropout_forward = None

        self.__observed_activation = observed_activation
        self.__input_gate_activation = input_gate_activation
        self.__forget_gate_activation = forget_gate_activation
        self.__output_gate_activation = output_gate_activation
        self.__hidden_activation = hidden_activation
        self.__output_activation = output_activation
        self.__output_layer_flag = output_layer_flag

        self.__computable_loss = computable_loss
        self.__learning_rate = learning_rate
        self.__hidden_n = hidden_n
        self.__output_n = output_n
        self.__dropout_rate = dropout_rate
        self.__input_adjusted_flag = input_adjusted_flag

        for v in regularizatable_data_list:
            if isinstance(v, RegularizatableData) is False:
                raise TypeError("The type of values of `regularizatable_data_list` must be `Regularizatable`.")
        self.__regularizatable_data_list = regularizatable_data_list

        self.__ctx = ctx

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        self.__input_dim = None
        self.__input_seq_len = None

        self.__output_layer_flag = output_layer_flag
        self.__output_no_bias_flag = output_no_bias_flag
        self.__output_nn = output_nn
        self.seq_len = seq_len

        self.epoch = 0
        self.__loss_list = []

    def initialize_params(self, input_dim, input_seq_len):
        '''
        Initialize params.

        Args:
            input_dim:      The number of units in input layer.
        '''
        if self.__input_dim is not None:
            return
        self.__input_dim = input_dim
        self.__input_seq_len = input_seq_len

        if self.__not_init_flag is False:
            if self.init_deferred_flag is False:
                self.observed_fc = nn.Linear(
                    input_dim,
                    self.__hidden_n * 4, 
                    bias=False, 
                )
                if self.initializer_f is None:
                    self.observed_fc.weight = torch.nn.init.xavier_normal_(
                        self.observed_fc.weight,
                        gain=1.0
                    )
                else:
                    self.observed_fc.weight = self.initializer_f(self.observed_fc.weight)

                self.hidden_fc = nn.Linear(
                    self.__hidden_n,
                    self.__hidden_n * 4, 
                )
                if self.initializer_f is None:
                    self.hidden_fc.weight = torch.nn.init.xavier_normal_(
                        self.hidden_fc.weight,
                        gain=1.0
                    )
                else:
                    self.hidden_fc.weight = self.initializer_f(self.observed_fc.weight)

                self.input_gate_fc = nn.Linear(
                    self.__hidden_n,
                    self.__hidden_n, 
                    bias=False, 
                )
                self.forget_gate_fc = nn.Linear(
                    self.__hidden_n,
                    self.__hidden_n, 
                    bias=False, 
                )
                self.output_gate_fc = nn.Linear(
                    self.__hidden_n,
                    self.__hidden_n, 
                    bias=False, 
                )

            self.output_fc = None
            self.output_nn = None
            if self.__output_layer_flag is True and self.__output_nn is None:
                if self.__output_no_bias_flag is True:
                    use_bias = False
                else:
                    use_bias = True

                # Different from mxnet version.
                self.output_fc = nn.Linear(
                    self.__hidden_n * self.__input_seq_len,
                    self.__output_n * self.__input_seq_len, 
                    bias=use_bias
                )
                self.__output_dim = self.__output_n
            elif self.__output_nn is not None:
                self.output_nn = self.__output_nn
                self.__output_dim = self.output_nn.units_list[-1]
            else:
                self.__output_dim = self.__hidden_n

        self.to(self.__ctx)
        if self.init_deferred_flag is False:
            if self.__not_init_flag is False:
                if self.optimizer_f is None:
                    self.optimizer = Adam(
                        self.parameters(), 
                        lr=self.__learning_rate,
                    )
                else:
                    self.optimizer = self.optimizer_f(
                        self.parameters(), 
                    )

    def learn(self, iteratable_data):
        '''
        Learn the observed data points
        for vector representation of the input time-series.

        Args:
            iteratable_data:     is-a `IteratableData`.

        '''
        if isinstance(iteratable_data, IteratableData) is False:
            raise TypeError("The type of `iteratable_data` must be `IteratableData`.")

        self.__loss_list = []
        learning_rate = self.__learning_rate
        try:
            epoch = self.epoch
            iter_n = 0
            for batch_observed_arr, batch_target_arr, test_batch_observed_arr, test_batch_target_arr in iteratable_data.generate_learned_samples():
                self.__batch_size = batch_observed_arr.shape[0]
                self.__seq_len = batch_observed_arr.shape[1]
                self.initialize_params(
                    input_dim=batch_observed_arr.reshape(
                        self.__batch_size * self.__seq_len,
                        -1
                    ).shape[-1],
                    input_seq_len=self.__seq_len
                )
                if self.output_nn is not None:
                    if hasattr(self.output_nn, "optimizer") is False:
                        _ = self.inference(batch_observed_arr)

                self.optimizer.zero_grad()
                if self.output_nn is not None:
                    self.output_nn.optimizer.zero_grad()

                # rank-3
                pred_arr = self.inference(batch_observed_arr)
                loss = self.compute_loss(
                    pred_arr,
                    batch_target_arr
                )
                loss.backward()

                if self.output_nn is not None:
                    self.output_nn.optimizer.step()
                self.optimizer.step()
                self.regularize()

                if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                    with torch.inference_mode():
                        # rank-3
                        test_pred_arr = self.inference(test_batch_observed_arr)

                        test_loss = self.compute_loss(
                            test_pred_arr,
                            test_batch_target_arr
                        )

                    _loss = loss.to('cpu').detach().numpy().copy()
                    _test_loss = test_loss.to('cpu').detach().numpy().copy()
                    self.__loss_list.append((_loss, _test_loss))
                    self.__logger.debug("Epochs: " + str(epoch + 1) + " Train loss: " + str(_loss) + " Test loss: " + str(_test_loss))
                    epoch += 1
                iter_n += 1

        except KeyboardInterrupt:
            self.__logger.debug("Interrupt.")

        self.epoch = epoch

        self.__logger.debug("end. ")

    def inference(self, observed_arr):
        '''
        Inference the feature points to reconstruct the time-series.

        Args:
            observed_arr:           rank-3 array like or sparse matrix as the observed data points.

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        return self(observed_arr)

    def compute_loss(self, pred_arr, labeled_arr):
        '''
        Compute loss.

        Args:
            pred_arr:       `mxnet.ndarray` or `mxnet.symbol`.
            labeled_arr:    `mxnet.ndarray` or `mxnet.symbol`.

        Returns:
            loss.
        '''
        return self.__computable_loss(pred_arr, labeled_arr)

    def regularize(self):
        '''
        Regularization.
        '''
        if len(self.__regularizatable_data_list) > 0:
            params_dict = self.extract_learned_dict()
            for regularizatable in self.__regularizatable_data_list:
                params_dict = regularizatable.regularize(params_dict)

            for k, params in params_dict.items():
                self.load_state_dict({k: params}, strict=False)

    def extract_learned_dict(self):
        '''
        Extract (pre-) learned parameters.

        Returns:
            `dict` of the parameters.
        '''
        params_dict = {}
        for k in self.state_dict().keys():
            params_dict.setdefault(k, self.state_dict()[k])

        return params_dict

    def extract_feature_points(self):
        '''
        Extract the activities in hidden layer and reset it, 
        considering this method will be called per one cycle in instances of time-series.

        Returns:
            The `mxnet.ndarray` of array like or sparse matrix of feature points or virtual visible observed data points.
        '''
        return self.feature_points_arr

    def forward(self, x):
        '''
        Forward with Gluon API.

        Args:
            x:      `mxnet.ndarray` of observed data points.
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        self.__batch_size = x.shape[0]
        self.__seq_len = x.shape[1]
        x = x.reshape(
            self.__batch_size,
            self.__seq_len,
            -1
        )
        self.initialize_params(
            input_dim=x.shape[2],
            input_seq_len=self.__seq_len
        )

        hidden_activity_arr = self.hidden_forward_propagate(x)

        if self.__dropout_rate > 0:
            hidden_activity_arr = self.dropout_forward(hidden_activity_arr)
        self.feature_points_arr = hidden_activity_arr

        if self.output_nn is not None:
            pred_arr = self.output_nn(hidden_activity_arr)
            return pred_arr
        if self.__output_layer_flag is True:
            # rank-3
            pred_arr = self.output_forward_propagate(hidden_activity_arr)
            return pred_arr
        else:
            return hidden_activity_arr

    def hidden_forward_propagate(self, observed_arr):
        '''
        Forward propagation in LSTM gate.

        Args:
            F:                      `mxnet.ndarray` or `mxnet.symbol`.
            observed_arr:           rank-3 tensor of observed data points.
        
        Returns:
            Predicted data points.
        '''
        pred_arr = None

        hidden_activity_arr = torch.zeros((self.__batch_size, self.__hidden_n), dtype=torch.float32)
        hidden_activity_arr = hidden_activity_arr.to(self.__ctx)
        cec_activity_arr = torch.zeros((self.__batch_size, self.__hidden_n), dtype=torch.float32)
        cec_activity_arr = cec_activity_arr.to(self.__ctx)

        if self.seq_len is not None:
            cycle_n = self.seq_len
        else:
            cycle_n = self.__seq_len

        for cycle in range(cycle_n):
            if cycle == 0:
                if observed_arr[:, cycle:cycle+1].shape[1] != 0:
                    hidden_activity_arr, cec_activity_arr = self.__lstm_forward(
                        observed_arr[:, cycle:cycle+1],
                        hidden_activity_arr,
                        cec_activity_arr
                    )
                    skip_flag = False
                else:
                    skip_flag = True
            else:
                if observed_arr.shape[1] > 1:
                    x_arr = observed_arr[:, cycle:cycle+1]
                else:
                    x_arr = torch.unsqueeze(pred_arr[:, -1], axis=1)

                if x_arr.shape[1] != 0:
                    hidden_activity_arr, cec_activity_arr = self.__lstm_forward(
                        x_arr,
                        hidden_activity_arr,
                        cec_activity_arr
                    )
                    skip_flag = False
                else:
                    skip_flag = True

            if skip_flag is False:
                add_arr = torch.unsqueeze(hidden_activity_arr, axis=1)
                if pred_arr is None:
                    pred_arr = add_arr
                else:
                    pred_arr = torch.cat((pred_arr, add_arr), dim=1)

        return pred_arr

    def __lstm_forward(
        self,
        observed_arr,
        hidden_activity_arr,
        cec_activity_arr
    ):
        '''
        Forward propagate in LSTM gate.
        
        Args:
            F:                      `mxnet.ndarray` or `mxnet.symbol`.
            observed_arr:           rank-2 tensor of observed data points.
            hidden_activity_arr:    rank-2 tensor of activities in hidden layer.
            cec_activity_arr:       rank-2 tensor of activities in the constant error carousel.
        
        Returns:
            Tuple data.
            - rank-2 tensor of activities in hidden layer,
            - rank-2 tensor of activities in LSTM gate.
        '''
        if len(observed_arr.shape) == 3:
            observed_arr = observed_arr[:, 0]

        if self.__input_adjusted_flag is True:
            observed_arr = torch.div(
                observed_arr, 
                torch.sum(torch.ones_like(observed_arr))
            )

        observed_lstm_matrix = self.observed_fc(observed_arr)

        # using bias
        hidden_lstm_matrix = self.hidden_fc(hidden_activity_arr)
        lstm_matrix = observed_lstm_matrix + hidden_lstm_matrix

        given_activity_arr = lstm_matrix[:, :self.__hidden_n]
        input_gate_activity_arr = lstm_matrix[:, self.__hidden_n:self.__hidden_n * 2]
        forget_gate_activity_arr = lstm_matrix[:, self.__hidden_n * 2:self.__hidden_n * 3]
        output_gate_activity_arr = lstm_matrix[:, self.__hidden_n * 3:self.__hidden_n * 4]

        # no bias
        _input_gate_activity_arr = self.input_gate_fc(cec_activity_arr)
        input_gate_activity_arr = input_gate_activity_arr + _input_gate_activity_arr
        # no bias
        _forget_gate_activity_arr = self.forget_gate_fc(cec_activity_arr)
        forget_gate_activity_arr = forget_gate_activity_arr + _forget_gate_activity_arr
        given_activity_arr = self.__observed_activation(given_activity_arr)
        input_gate_activity_arr = self.__input_gate_activation(input_gate_activity_arr)
        forget_gate_activity_arr = self.__forget_gate_activation(forget_gate_activity_arr)

        # rank-2
        _cec_activity_arr = torch.mul(given_activity_arr, input_gate_activity_arr) + torch.mul(forget_gate_activity_arr, cec_activity_arr)

        # no bias
        _output_gate_activity_arr = self.output_gate_fc(_cec_activity_arr)

        output_gate_activity_arr = output_gate_activity_arr + _output_gate_activity_arr
        output_gate_activity_arr = self.__output_gate_activation(output_gate_activity_arr)

        # rank-2
        _hidden_activity_arr = torch.mul(
            output_gate_activity_arr, 
            self.__hidden_activation(_cec_activity_arr)
        )

        return (_hidden_activity_arr, _cec_activity_arr)

    def output_forward_propagate(self, pred_arr):
        '''
        Forward propagation in output layer.
        
        Args:
            F:                   `mxnet.ndarray` or `mxnet.symbol`.
            pred_arr:            rank-3 tensor of predicted data points.

        Returns:
            rank-3 tensor of propagated data points.
        '''
        if self.__output_layer_flag is False:
            return pred_arr

        batch_size = pred_arr.shape[0]
        seq_len = pred_arr.shape[1]
        # Different from mxnet version.
        pred_arr = self.output_fc(
            torch.reshape(
                pred_arr, 
                (batch_size, -1)
            )
        )
        if self.__output_activation == "identity_adjusted":
            pred_arr = torch.div(pred_arr, torch.sum(torch.ones_like(pred_arr)))
        elif self.__output_activation != "identity":
            pred_arr = self.__output_activation(pred_arr)
        pred_arr = torch.reshape(
            pred_arr, 
            (batch_size, seq_len, -1)
        )
        return pred_arr

    def save_parameters(self, filename):
        '''
        Save parameters to files.

        Args:
            filename:       File name.
        '''
        torch.save(
            {
                'epoch': self.epoch,
                'model_state_dict': self.state_dict(),
                'optimizer_state_dict': self.optimizer.state_dict(),
                'loss': self.loss_arr,
                'input_dim': self.__input_dim,
                'input_seq_len': self.__input_seq_len,
            }, 
            filename
        )

    def load_parameters(self, filename, ctx=None, strict=True):
        '''
        Load parameters to files.

        Args:
            filename:       File name.
            ctx:            Context-manager that changes the selected device.
            strict:         Whether to strictly enforce that the keys in state_dict match the keys returned by this module’s state_dict() function. Default: `True`.
        '''
        checkpoint = torch.load(filename)
        self.initialize_params(
            input_dim=checkpoint["input_dim"],
            input_seq_len=checkpoint["input_seq_len"],
        )
        self.load_state_dict(checkpoint['model_state_dict'], strict=strict)
        self.optimizer.load_state_dict(
            checkpoint['optimizer_state_dict']
        )
        self.epoch = checkpoint['epoch']
        self.__loss_list = checkpoint['loss'].tolist()
        if ctx is not None:
            self.to(ctx)
            self.__ctx = ctx

    def set_readonly(self, value):
        ''' setter for losses. '''
        raise TypeError("This property must be read-only.")

    __loss_list = []

    def get_loss_arr(self):
        ''' getter for losses. '''
        return np.array(self.__loss_list)

    loss_arr = property(get_loss_arr, set_readonly)

    def get_output_dim(self):
        return self.__output_dim

    output_dim = property(get_output_dim, set_readonly)

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not.'''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not. '''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)



# -*- coding: utf-8 -*-
from mxnet.initializer import Initializer
import mxnet.ndarray as nd
from mxnet import gluon
import numpy as np


class NormalInitializer(Initializer):
    '''
    Initializes weights with the probability density function of the normal distribution.
    '''
    
    def __init__(self, mu=0.0, sigma=1.0, scale=1.0, init_dtype=np.float16):
        '''
        Init.

        Args:
            mu:                 Mean (“centre”) of the distribution.
            sigma:              Standard deviation (spread or “width”) of the distribution.
            init_dtype:         `np.dtype`.
            scale:              `float` of scaling factor for initial parmaeters.
        '''

        super(NormalInitializer, self).__init__(sigma=sigma, mu=mu, init_dtype=init_dtype)
        self.__sigma = sigma
        self.__mu = mu
        self.__init_dtype = init_dtype
        self.__scale = scale

    def _init_weight(self, name, arr):
        nd.random.normal(self.__mu, self.__sigma, out=arr)
        arr = arr.astype(self.__init_dtype)
        arr = arr * self.__scale



# -*- coding: utf-8 -*-
import torch
import torch.nn as nn
import numpy as np
from accelbrainbase.computable_loss import ComputableLoss
from accelbrainbase.computableloss._torch.l2_norm_loss import L2NormLoss


class DRCNLoss(nn.modules.loss._Loss, ComputableLoss):
    '''
    Loss function of Deep Reconstruction-Classification Networks.

    Deep Reconstruction-Classification Network(DRCN) is a convolutional network 
    that jointly learns two tasks: 
    
    1. supervised source label prediction.
    2. unsupervised target data reconstruction. 

    Ideally, a discriminative representation should model both the label and 
    the structure of the data. Based on that intuition, Ghifary, M., et al.(2016) hypothesize 
    that a domain-adaptive representation should satisfy two criteria:
    
    1. classify well the source domain labeled data.
    2. reconstruct well the target domain unlabeled data, which can be viewed as an approximate of the ideal discriminative representation.

    The encoding parameters of the DRCN are shared across both tasks, 
    while the decoding parameters are sepa-rated. The aim is that the learned label 
    prediction function can perform well onclassifying images in the target domain
    thus the data reconstruction can beviewed as an auxiliary task to support the 
    adaptation of the label prediction.

    References:
        - Ghifary, M., Kleijn, W. B., Zhang, M., Balduzzi, D., & Li, W. (2016, October). Deep reconstruction-classification networks for unsupervised domain adaptation. In European Conference on Computer Vision (pp. 597-613). Springer, Cham.
    '''

    def __init__(
        self, 
        axis=-1,
        rc_lambda=0.75,
        weight=1.0,
        label_smoothing=0.0,
        classification_weight=None, 
        reconstruction_weight=None, 
        batch_axis=1, 
    ):
        '''
        Init.

        Args:
            axis:                   The axis to sum over when computing softmax and entropy.
            sparse_label:           Whether label is an integer array instead of probability distribution.
            rc_lambda:              Tradeoff parameter for loss functions.
            from_logits:            Whether input is a log probability (usually from log_softmax) instead of unnormalized numbers.
            weight:                 Global scalar weight for total loss.
            classification_weight:  Global scalar weight for classification loss. If `None`, this value will be equivalent to `weight`.
            reconstruction_weight:  Global scalar weight for reconstruction loss. If `None`, this value will be equivalent to `weight`.
            batch_axis:             The axisthat represents mini-batch.
        '''
        super(DRCNLoss, self).__init__()
        self.__cross_entropy_loss = torch.nn.CrossEntropyLoss(
            size_average=None, 
            ignore_index=-100, 
            reduce=None, 
            reduction='mean', 
            label_smoothing=label_smoothing
        )
        self.__l2_norm_loss = L2NormLoss(batch_axis=batch_axis)
        self.__classification_weight = classification_weight
        self.__reconstruction_weight = reconstruction_weight
        self.__weight = weight
        self.__rc_lambda = rc_lambda

    def compute(
        self, 
        decoded_arr, 
        pred_arr, 
        observed_arr, 
        label_arr, 
    ):
        '''
        Compute loss.

        Args:
            decoded_arr:            `mxnet.ndarray` or `mxnet.symbol` of decoded feature points.
            pred_arr:               `mxnet.ndarray` or `mxnet.symbol` of inferenced labeled feature points.
            observed_arr:           `mxnet.ndarray` or `mxnet.symbol` of observed data points.
            label_arr:              `mxnet.ndarray` or `mxnet.symbol` of label data.

        Returns:
            Tensor of losses.
        '''
        return self(
            decoded_arr, 
            pred_arr, 
            observed_arr, 
            label_arr, 
        )

    def forward(
        self, 
        decoded_arr, 
        pred_arr, 
        observed_arr, 
        label_arr, 
    ):
        '''
        Forward propagation, computing losses.

        Args:
            decoded_arr:            `mxnet.ndarray` or `mxnet.symbol` of decoded feature points.
            pred_arr:               `mxnet.ndarray` or `mxnet.symbol` of inferenced labeled feature points.
            observed_arr:           `mxnet.ndarray` or `mxnet.symbol` of observed data points.
            label_arr:              `mxnet.ndarray` or `mxnet.symbol` of label data.

        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of loss.
        '''
        classification_loss = self.__cross_entropy_loss(
            pred_arr,
            label_arr
        )

        if self.__classification_weight is not None:
            classification_loss = classification_loss * self.__classification_weight

        observed_arr = observed_arr.reshape_as(decoded_arr)
        reconstruction_loss = self.__l2_norm_loss(observed_arr, decoded_arr)

        if self.__reconstruction_weight is not None:
            reconstruction_loss = reconstruction_loss * self.__reconstruction_weight

        classification_loss = classification_loss * self.__rc_lambda
        reconstruction_loss = reconstruction_loss * (1 - self.__rc_lambda)
        if self.__weight is not None:
            classification_loss = classification_loss * self.__weight
            reconstruction_loss = reconstruction_loss * self.__weight

        return classification_loss + reconstruction_loss, classification_loss, reconstruction_loss



# -*- coding: utf-8 -*-
from accelbrainbase.observabledata._mxnet.transformer_model import TransformerModel
from accelbrainbase.observabledata._mxnet.attention_model import AttentionModel
from accelbrainbase.observabledata._mxnet.attentionmodel.multi_head_attention_model import MultiHeadAttentionModel
from accelbrainbase.observabledata._mxnet.attentionmodel.multiheadattentionmodel.self_attention_model import SelfAttentionModel
from accelbrainbase.computable_loss import ComputableLoss
from accelbrainbase.regularizatable_data import RegularizatableData

from mxnet.gluon.block import HybridBlock
from mxnet import gluon
from mxnet import autograd
import numpy as np
import mxnet as mx
import mxnet.ndarray as nd
from mxnet import MXNetError
from logging import getLogger


class TransformerReconstructor(HybridBlock, TransformerModel):
    '''
    Reconstructor of Transformering Auto-Encoder.

    References:
        - Bahdanau, D., Cho, K., & Bengio, Y. (2014). Neural machine translation by jointly learning to align and translate. arXiv preprint arXiv:1409.0473.
        - Floridi, L., & Chiriatti, M. (2020). GPT-3: Its nature, scope, limits, and consequences. Minds and Machines, 30(4), 681-694.
        - Miller, A., Fisch, A., Dodge, J., Karimi, A. H., Bordes, A., & Weston, J. (2016). Key-value memory networks for directly reading documents. arXiv preprint arXiv:1606.03126.
        - Radford, A., Narasimhan, K., Salimans, T., & Sutskever, I. (2018) Improving Language Understanding by Generative Pre-Training. OpenAI (URL: https://s3-us-west-2.amazonaws.com/openai-assets/research-covers/language-unsupervised/language_understanding_paper.pdf)
        - Radford, A., Wu, J., Child, R., Luan, D., Amodei, D., & Sutskever, I. (2019). Language models are unsupervised multitask learners. OpenAI blog, 1(8), 9.
        - Vaswani, A., Shazeer, N., Parmar, N., Uszkoreit, J., Jones, L., Gomez, A. N., & Polosukhin, I. (2017). Attention is all you need. arXiv preprint arXiv:1706.03762.

    '''

    __batch_size = None

    def get_batch_size(self):
        ''' getter '''
        return self.__batch_size
    
    def set_batch_size(self, value):
        ''' setter '''
        self.__batch_size = value
        for i in range(len(self.self_attention_list)):
            self.self_attention_list[i].batch_size = value

    batch_size = property(get_batch_size, set_batch_size)

    __head_n = None

    def get_head_n(self):
        ''' getter '''
        return self.__head_n
    
    def set_head_n(self, value):
        ''' setter '''
        self.__head_n = value
        for i in range(len(self.self_attention_list)):
            self.self_attention_list[i].head_n = value

    head_n = property(get_head_n, set_head_n)

    __seq_len = None

    def get_seq_len(self):
        ''' getter '''
        return self.__seq_len
    
    def set_seq_len(self, value):
        ''' setter '''
        self.__seq_len = value
        for i in range(len(self.self_attention_list)):
            self.self_attention_list[i].seq_len = value

    seq_len = property(get_seq_len, set_seq_len)

    __depth_dim = None

    def get_depth_dim(self):
        ''' getter '''
        return self.__depth_dim
    
    def set_depth_dim(self, value):
        ''' setter '''
        self.__depth_dim = value
        for i in range(len(self.self_attention_list)):
            self.self_attention_list[i].depth_dim = value

    depth_dim = property(get_depth_dim, set_depth_dim)

    def __init__(
        self,
        depth_dim,
        layer_n,
        computable_loss,
        head_n=3,
        initializer=None,
        self_attention_layer_norm_list=[],
        fc_layer_norm_list=[],
        filter_fc_list=[],
        dropout_rate=0.1,
        fc_list=[],
        self_attention_activation_list=[],
        fc_activation_list=[],
        output_layer_norm=None,
        output_fc=None,
        output_dim=None,
        optimizer_name="SGD",
        learning_rate=1e-03,
        learning_attenuate_rate=1.0,
        attenuate_epoch=50,
        regularizatable_data_list=[],
        ctx=mx.gpu(),
        hybridize_flag=True,
        not_init_flag=False,
        **kwargs
    ):
        """
        Init.

        Args:
            computable_loss:                is-a `ComputableLoss` or `gluon.loss`.
            encoder:                        is-a `TransformerModel`.
            decoder:                        is-a `TransformerModel`.
            layer_n:                        `int` of the number of layers.
            head_n:                         `int` of the number of heads for multi-head attention model.
            seq_len:                        `int` of the length of sequences.
            depth_dim:                      `int` of dimension of dense layer.
            hidden_dim:                     `int` of dimension of hidden(encoder) layer.
            self_attention_activation_list: `list` of `str` of activation function for self-attention model.
            fc_activation_list:             `list` of `str` of activation function in fully-connected layers.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
            optimizer_name:                 `str` of name of optimizer.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            scale:                          `float` of scaling factor for initial parameters.
            ctx:                            `mx.cpu()` or `mx.gpu()`.
            initializer:                    is-a `mxnet.initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.

        """
        if isinstance(depth_dim, int) is False:
            raise TypeError("The type of `depth_dim` must be `int`.")
        if isinstance(computable_loss, ComputableLoss) is False and isinstance(computable_loss, gluon.loss.Loss) is False:
            raise TypeError("The type of `computable_loss` must be `ComputableLoss` or `gluon.loss.Loss`.")

        super(TransformerReconstructor, self).__init__()
        self.__computable_loss = computable_loss

        if initializer is None:
            self.initializer = mx.initializer.Normal(1.0)
        else:
            if isinstance(initializer, mx.initializer.Initializer) is False:
                raise TypeError("The type of `initializer` must be `mxnet.initializer.Initializer`.")

            self.initializer = initializer

        self.layer_n = layer_n

        self.self_attention_activation_list = self_attention_activation_list
        if len(self.self_attention_activation_list) != layer_n:
            self.self_attention_activation_list = self.self_attention_activation_list * layer_n

        self.fc_activation_list = fc_activation_list
        if len(self.fc_activation_list) != layer_n:
            self.fc_activation_list = self.fc_activation_list * layer_n

        with self.name_scope():
            self.embedding_dropout = gluon.nn.Dropout(rate=dropout_rate)

            if len(self_attention_layer_norm_list) == 0:
                self.self_attention_layer_norm_list = [None] * layer_n
            else:
                self.self_attention_layer_norm_list = self_attention_layer_norm_list

            self.self_attention_list = [None] * layer_n

            if len(filter_fc_list) == 0:
                self.filter_fc_list = [None] * layer_n
            else:
                self.filter_fc_list = filter_fc_list

            self.self_attention_dropout_list = [None] * layer_n

            if len(fc_layer_norm_list) == 0:
                self.fc_layer_norm_list = [None] * layer_n
            else:
                self.fc_layer_norm_list = fc_layer_norm_list

            self.fc_dropout_list = [None] * layer_n

            if len(fc_list) == 0:
                self.fc_list = [None] * layer_n
            else:
                self.fc_list = fc_list

            for i in range(layer_n):
                if self.self_attention_layer_norm_list[i] is None:
                    self.self_attention_layer_norm_list[i] = gluon.nn.LayerNorm()
                self.register_child(self.self_attention_layer_norm_list[i])
                self.self_attention_list[i] = SelfAttentionModel(
                    depth_dim=depth_dim,
                    computable_loss=computable_loss,
                    not_init_flag=not_init_flag,
                )
                self.self_attention_list[i].head_n = head_n
                self.register_child(self.self_attention_list[i])

                self.self_attention_dropout_list[i] = gluon.nn.Dropout(rate=dropout_rate)
                self.register_child(self.self_attention_dropout_list[i])

                if self.fc_layer_norm_list[i] is None:
                    self.fc_layer_norm_list[i] = gluon.nn.LayerNorm()
                self.register_child(self.fc_layer_norm_list[i])

                if self.filter_fc_list[i] is None:
                    self.filter_fc_list[i] = gluon.nn.Dense(
                        depth_dim, 
                        use_bias=True,
                        flatten=False,
                    )
                self.register_child(self.filter_fc_list[i])

                if self.fc_list[i] is None:
                    self.fc_list[i] = gluon.nn.Dense(
                        depth_dim, 
                        use_bias=True,
                        flatten=False,
                    )
                self.register_child(self.fc_list[i])

                self.fc_dropout_list[i] = gluon.nn.Dropout(rate=dropout_rate)
                self.register_child(self.fc_dropout_list[i])

                if self.self_attention_activation_list[i] == "GELU" or self.self_attention_activation_list[i] == "gelu":
                    self.self_attention_activation_list[i] = mx.gluon.nn.GELU()
                    self.register_child(self.self_attention_activation_list[i])

                if self.fc_activation_list[i] == "GELU" or self.fc_activation_list[i] == "gelu":
                    self.fc_activation_list[i] = mx.gluon.nn.GELU()
                    self.register_child(self.fc_activation_list[i])

            if output_layer_norm is None:
                self.output_layer_norm = gluon.nn.LayerNorm()
            else:
                self.output_layer_norm = output_layer_norm
            self.register_child(self.output_layer_norm)

            if output_fc is None:
                if output_dim is None:
                    output_dim = depth_dim
                self.output_fc = gluon.nn.Dense(
                    output_dim,
                    use_bias=False,
                    flatten=False,
                )
            else:
                self.output_fc = output_fc
            self.register_child(self.output_fc)

        if self.init_deferred_flag is False:
            if not_init_flag is False:
                self.collect_params().initialize(self.initializer, force_reinit=True, ctx=ctx)
                self.trainer = gluon.Trainer(self.collect_params(), optimizer_name, {"learning_rate": learning_rate})
                if hybridize_flag is True:
                    self.hybridize()

        self.learning_rate = learning_rate
        self.learning_attenuate_rate = learning_attenuate_rate
        self.attenuate_epoch = attenuate_epoch

        for v in regularizatable_data_list:
            if isinstance(v, RegularizatableData) is False:
                raise TypeError("The type of values of `regularizatable_data_list` must be `RegularizatableData`.")
        self.__regularizatable_data_list = regularizatable_data_list

        self.__ctx = ctx

        logger = getLogger("accelbrainbase")
        self.logger = logger

        self.head_n = head_n
        self.depth_dim = depth_dim

    def learn(self, iteratable_data):
        '''
        Learn samples drawn by `IteratableData.generate_learned_samples()`.

        Args:
            iteratable_data:     is-a `IteratableData`.

        '''
        raise NotImplementedError("This class itself can do learning. Use `TransformerController`.")

    def inference(
        self, 
        observed_arr, 
        mask=None, 
    ):
        '''
        Inference samples drawn by `IteratableData.generate_inferenced_samples()`.

        Args:
            observed_arr:                   rank-3 Array like or sparse matrix as the observed data points.
                                            The shape is: (batch size, the length of sequence, feature points)

            mask:       rank-3 Array like or sparse matrix as the mask.
                                           The shape is: (batch size, the length of sequence, feature points)

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        observed_arr = self.embedding(observed_arr)

        if mask is None:
            mask = nd.ones(
                shape=(
                    observed_arr.shape[0],
                    1,
                    1,
                    1
                ),
                ctx=observed_arr.context
            )

        return self(
            observed_arr, 
            mask, 
        )

    def hybrid_forward(
        self, 
        F,
        observed_arr, 
        mask=None, 
    ):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:                              `mxnet.ndarray` or `mxnet.symbol`.
            observed_arr:                   rank-3 Array like or sparse matrix as the observed data points.
                                            The shape is: (batch size, the length of sequence, feature points)

            mask:       rank-3 Array like or sparse matrix as the mask.
                                           The shape is: (batch size, the length of sequence, feature points)
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        # rank-3
        return self.forward_propagation(
            F, 
            observed_arr, 
            mask, 
        )

    def forward_propagation(
        self, 
        F,
        observed_arr, 
        mask=None, 
    ):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:                              `mxnet.ndarray` or `mxnet.symbol`.
            observed_arr:                   rank-3 Array like or sparse matrix as the observed data points.
                                            The shape is: (batch size, the length of sequence, feature points)

            mask:       rank-3 Array like or sparse matrix as the mask.
                                           The shape is: (batch size, the length of sequence, feature points)
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        if self.embedding_flag is True:
            x = self.embedding_dropout(observed_arr)
        else:
            x = observed_arr

        for i in range(self.layer_n):
            _x = self.self_attention_layer_norm_list[i](x)
            _x = self.self_attention_list[i].inference(_x, mask)
            _x = self.self_attention_dropout_list[i](_x)

            if len(self.self_attention_activation_list) > 0:
                if isinstance(self.self_attention_activation_list[i], gluon.nn.activations.GELU) is True:
                    _x = self.self_attention_activation_list[i](_x)
                elif self.self_attention_activation_list[i] == "identity_adjusted":
                    _x = _x / F.sum(F.ones_like(_x))
                elif self.self_attention_activation_list[i] == "softmax":
                    _x = F.softmax(_x)
                elif self.self_attention_activation_list[i] == "log_softmax":
                    _x = F.softmax(_x)
                elif self.self_attention_activation_list[i] != "identity":
                    _x = F.Activation(_x, self.self_attention_activation_list[i])

            x = x + _x

            _x = self.fc_layer_norm_list[i](x)
            _x = self.filter_fc_list[i](_x)
            _x = self.fc_dropout_list[i](_x)
            _x = self.fc_list[i](_x)

            if len(self.fc_activation_list) > 0:
                if isinstance(self.fc_activation_list[i], gluon.nn.activations.GELU) is True:
                    _x = self.fc_activation_list[i](_x)
                elif self.fc_activation_list[i] == "identity_adjusted":
                    _x = _x / F.sum(F.ones_like(_x))
                elif self.fc_activation_list[i] == "softmax":
                    _x = F.softmax(_x)
                elif self.fc_activation_list[i] == "log_softmax":
                    _x = F.softmax(_x)
                elif self.fc_activation_list[i] != "identity":
                    _x = F.Activation(_x, self.fc_activation_list[i])

            x = x + _x

        y = self.output_layer_norm(x)
        y = self.output_fc(y)

        return y



# -*- coding: utf-8 -*-
from accelbrainbase.iteratable_data import IteratableData
from accelbrainbase.regularizatable_data import RegularizatableData
from accelbrainbase.observabledata._torch.neural_networks import NeuralNetworks
from accelbrainbase.observable_data import ObservableData
from accelbrainbase.computable_loss import ComputableLoss

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.optim.optimizer import Optimizer, required
from torch.optim.adam import Adam

import numpy as np
from logging import getLogger


class ConvolutionalNeuralNetworks(nn.Module, ObservableData):
    '''
    Convolutional Neural Networks.

    References:
        - Dumoulin, V., & V,kisin, F. (2016). A guide to convolution arithmetic for deep learning. arXiv preprint arXiv:1603.07285.
        - Kamyshanska, H., & Memisevic, R. (2014). The potential energy of an autoencoder. IEEE transactions on pattern analysis and machine intelligence, 37(6), 1261-1273.
        - Masci, J., Meier, U., Cireşan, D., & Schmidhuber, J. (2011, June). Stacked convolutional auto-encoders for hierarchical feature extraction. In International Conference on Artificial Neural Networks (pp. 52-59). Springer, Berlin, Heidelberg.
    '''

    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    # is-a `NNHybrid`.
    __input_nn = None
    # is-a `NNHybrid`.
    __output_nn = None

    # `bool`, compute accuracy or not in learning.
    __compute_acc_flag = True

    def get_compute_acc_flag(self):
        ''' getter '''
        return self.__compute_acc_flag
    
    def set_compute_acc_flag(self, value):
        ''' setter '''
        self.__compute_acc_flag = value

    compute_acc_flag = property(get_compute_acc_flag, set_compute_acc_flag)

    def __init__(
        self,
        computable_loss,
        initializer_f=None,
        optimizer_f=None,
        learning_rate=1e-05,
        hidden_units_list=[
            torch.nn.Conv2d(
                in_channels=3,
                out_channels=16,
                kernel_size=3,
                stride=1,
                padding=1,
            ), 
            torch.nn.Conv2d(
                in_channels=16,
                out_channels=32,
                kernel_size=3,
                stride=1,
                padding=1,
            ),
        ],
        input_nn=None,
        input_result_height=None,
        input_result_width=None,
        input_result_channel=None,
        output_nn=None,
        hidden_dropout_rate_list=[0.5, 0.5],
        hidden_batch_norm_list=[
            torch.nn.BatchNorm2d(16), 
            torch.nn.BatchNorm2d(32),
        ],
        hidden_activation_list=[
            torch.nn.functional.tanh, 
            torch.nn.functional.sigmoid
        ],
        hidden_residual_flag=False,
        hidden_dense_flag=False,
        dense_axis=1,
        ctx="cpu",
        regularizatable_data_list=[],
        scale=1.0,
        not_init_flag=False,
    ):
        '''
        Init.

        Args:
            computable_loss:                is-a `ComputableLoss` or `mxnet.gluon.loss`.
            initializer:                    is-a `mxnet.initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            learning_rate:                  `float` of learning rate.
            hidden_units_list:              `list` of `mxnet.gluon.nn._conv` in hidden layers.
            input_nn:                       is-a `NeuralNetworks` as input layers.
                                            If `None`, first layer in `hidden_units_list` will be considered as as input layer.

            output_nn:                      is-a `NeuralNetworks` as output layers.
                                            If `None`, last layer in `hidden_units_list` will be considered as an output layer.

            hidden_dropout_rate_list:       `list` of `float` of dropout rate in hidden layers.
            hidden_batch_norm_list:         `list` of `mxnet.gluon.nn.BatchNorm` in hidden layers.

            optimizer_name:                 `str` of name of optimizer.

            hidden_activation_list:         `list` of act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in input gate.

            hidden_residual_flag:           `bool` whether execute the residual learning or not in hidden layers.

            ctx:                            `mx.cpu()` or `mx.gpu()`.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            regularizatable_data_list:      `list` of `RegularizatableData`.
            scale:                          `float` of scaling factor for initial parameters.
            not_init_flag:                  `bool` of whether initialize parameters or not.
        '''
        if len(hidden_units_list) != len(hidden_activation_list):
            raise ValueError("The length of `hidden_units_list` and `hidden_activation_list` must be equivalent.")

        if len(hidden_dropout_rate_list) != len(hidden_units_list):
            raise ValueError("The length of `hidden_dropout_rate_list` and `hidden_units_list` must be equivalent.")

        if len(hidden_batch_norm_list) != len(hidden_units_list):
            raise ValueError("The length of `hidden_batch_norm_list` and `hidden_units_list` must be equivalent.")

        if isinstance(computable_loss, ComputableLoss) is False and isinstance(computable_loss, nn.modules.loss._Loss) is False:
            raise TypeError("The type of `computable_loss` must be `ComputableLoss` or `nn.modules.loss._Loss`.")

        super(ConvolutionalNeuralNetworks, self).__init__()

        self.input_nn = input_nn
        self.output_nn = output_nn

        self.hidden_units_list = nn.ModuleList(hidden_units_list)
        self.hidden_batch_norm_list = nn.ModuleList(hidden_batch_norm_list)

        dropout_list = [None] * len(hidden_dropout_rate_list)
        for i in range(len(hidden_dropout_rate_list)):
            dropout_list[i] = nn.Dropout(p=hidden_dropout_rate_list[i])
        self.hidden_dropout_rate_list = nn.ModuleList(dropout_list)

        self.__not_init_flag = not_init_flag

        self.__computable_loss = computable_loss
        self.__learning_rate = learning_rate

        for v in regularizatable_data_list:
            if isinstance(v, RegularizatableData) is False:
                raise TypeError("The type of values of `regularizatable_data_list` must be `RegularizatableData`.")
        self.__regularizatable_data_list = regularizatable_data_list

        for i in range(len(hidden_units_list)):
            if initializer_f is None:
                hidden_units_list[i].weight = torch.nn.init.xavier_normal_(
                    hidden_units_list[i].weight,
                    gain=1.0
                )
            else:
                hidden_units_list[i].weight = initializer_f(hidden_units_list[i].weight)

        self.hidden_units_list = nn.ModuleList(hidden_units_list)

        self.optimizer_f = optimizer_f
        self.__ctx = ctx

        self.to(self.__ctx)

        if self.init_deferred_flag is False:
            if self.__not_init_flag is False:
                if self.optimizer_f is None:
                    self.optimizer = Adam(
                        self.parameters(), 
                        lr=self.__learning_rate,
                    )
                else:
                    self.optimizer = self.optimizer_f(
                        self.parameters(), 
                    )

        self.flatten = nn.Flatten()

        self.hidden_activation_list = hidden_activation_list

        self.__computable_loss = computable_loss
        self.__learning_rate = learning_rate

        self.input_result_height = input_result_height
        self.input_result_width = input_result_width
        self.input_result_channel = input_result_channel

        self.__hidden_residual_flag = hidden_residual_flag
        self.__hidden_dense_flag = hidden_dense_flag
        self.__dense_axis = dense_axis

        self.__safe_params_dict = {}

        self.epoch = 0
        self.__loss_list = []
        logger = getLogger("accelbrainbase")
        self.__logger = logger

    def learn(self, iteratable_data):
        '''
        Learn the observed data points
        for vector representation of the input images.

        Args:
            iteratable_data:     is-a `IteratableData`.

        '''
        if isinstance(iteratable_data, IteratableData) is False:
            raise TypeError("The type of `iteratable_data` must be `IteratableData`.")

        self.__loss_list = []
        self.__acc_list = []
        learning_rate = self.__learning_rate
        try:
            epoch = self.epoch
            iter_n = 0
            for batch_observed_arr, batch_target_arr, test_batch_observed_arr, test_batch_target_arr in iteratable_data.generate_learned_samples():
                self.epoch = epoch
                self.batch_size = batch_observed_arr.shape[0]

                if self.output_nn is not None:
                    if hasattr(self.output_nn, "optimizer") is False:
                        _ = self.inference(batch_observed_arr)

                self.optimizer.zero_grad()
                if self.output_nn is not None:
                    self.output_nn.optimizer.zero_grad()

                # rank-3
                pred_arr = self.inference(batch_observed_arr)
                loss = self.compute_loss(
                    pred_arr,
                    batch_target_arr
                )
                loss.backward()
                if self.output_nn is not None:
                    self.output_nn.optimizer.step()
                self.optimizer.step()
                self.regularize()

                if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                    with torch.inference_mode():
                        # rank-3
                        test_pred_arr = self.inference(test_batch_observed_arr)

                        test_loss = self.compute_loss(
                            test_pred_arr,
                            test_batch_target_arr
                        )

                    _loss = loss.to('cpu').detach().numpy().copy()
                    _test_loss = test_loss.to('cpu').detach().numpy().copy()

                    self.__logger.debug("Epochs: " + str(epoch + 1) + " Train loss: " + str(loss.to('cpu').detach().numpy().mean()) + " Test loss: " + str(test_loss.to('cpu').detach().numpy().mean()))
                    self.__loss_list.append((_loss, _test_loss))

                if self.compute_acc_flag is True:
                    if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                        acc, inferenced_label_arr, answer_label_arr = self.compute_acc(pred_arr, batch_target_arr)
                        test_acc, test_inferenced_label_arr, test_answer_label_arr = self.compute_acc(test_pred_arr, test_batch_target_arr)

                        if ((epoch + 1) % 100 == 0):
                            self.__logger.debug("-" * 100)
                            self.__logger.debug("Train accuracy: " + str(acc) + " Test accuracy: " + str(test_acc))
                            self.__logger.debug("Train infenreced label(inferenced):")
                            self.__logger.debug(inferenced_label_arr.to('cpu').detach().numpy())
                            self.__logger.debug("Train infenreced label(answer):")
                            self.__logger.debug(answer_label_arr.to('cpu').detach().numpy())

                            self.__logger.debug("Test infenreced label(inferenced):")
                            self.__logger.debug(test_inferenced_label_arr.to('cpu').detach().numpy())
                            self.__logger.debug("Test infenreced label(answer):")
                            self.__logger.debug(test_answer_label_arr.to('cpu').detach().numpy())
                            self.__logger.debug("-" * 100)

                            if (test_answer_label_arr[0].to('cpu').detach().numpy() == test_answer_label_arr.to('cpu').detach().numpy()).astype(int).sum() != test_answer_label_arr.shape[0]:
                                if (test_inferenced_label_arr[0].to('cpu').detach().numpy() == test_inferenced_label_arr.to('cpu').detach().numpy()).astype(int).sum() == test_inferenced_label_arr.shape[0]:
                                    self.__logger.debug("It may be overfitting.")

                        self.__acc_list.append(
                            (
                                acc,
                                test_acc
                            )
                        )

                if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                    epoch += 1
                iter_n += 1

        except KeyboardInterrupt:
            self.__logger.debug("Interrupt.")

        self.__logger.debug("end. ")
        self.epoch = epoch

    def inference(self, observed_arr):
        '''
        Inference the labels.

        Args:
            observed_arr:   rank-4 Array like or sparse matrix as the observed data points.
                            The shape is: (batch size, channel, height, width)

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        return self(observed_arr)

    def compute_loss(self, pred_arr, labeled_arr):
        '''
        Compute loss.

        Args:
            pred_arr:       `mxnet.ndarray` or `mxnet.symbol`.
            labeled_arr:    `mxnet.ndarray` or `mxnet.symbol`.

        Returns:
            loss.
        '''
        return self.__computable_loss(pred_arr, labeled_arr)

    def regularize(self):
        '''
        Regularization.
        '''
        if len(self.__regularizatable_data_list) > 0:
            params_dict = self.extract_learned_dict()
            for regularizatable in self.__regularizatable_data_list:
                params_dict = regularizatable.regularize(params_dict)

            for k, params in params_dict.items():
                self.load_state_dict({k: params}, strict=False)

    def extract_learned_dict(self):
        '''
        Extract (pre-) learned parameters.

        Returns:
            `dict` of the parameters.
        '''
        params_dict = {}
        for k in self.state_dict().keys():
            params_dict.setdefault(k, self.state_dict()[k])

        return params_dict

    def forward(self, x):
        '''
        Hybrid forward with Gluon API.

        Args:
            x:      `tensor` of observed data points.
        
        Returns:
            `tensor` of inferenced feature points.
        '''
        if self.input_nn is not None:
            x = self.input_nn(x)
            x = torch.reshape(
                x, 
                shape=(
                    -1, 
                    self.input_result_channel, 
                    self.input_result_height, 
                    self.input_result_width
                )
            )

        for i in range(len(self.hidden_activation_list)):
            if i > 0 and i + 1 < len(self.hidden_activation_list):
                if self.__hidden_residual_flag is True or self.__hidden_dense_flag is True:
                    _x = x

            x = self.hidden_units_list[i](x)
            if self.hidden_activation_list[i] == "identity_adjusted":
                x = x / torch.sum(torch.ones_like(x))
            elif self.hidden_activation_list[i] != "identity":
                x = self.hidden_activation_list[i](x)

            if self.hidden_dropout_rate_list[i] is not None:
                x = self.hidden_dropout_rate_list[i](x)

            if self.hidden_batch_norm_list[i] is not None:
                x = self.hidden_batch_norm_list[i](x)

            if i > 0 and i + 1 < len(self.hidden_activation_list):
                if self.__hidden_residual_flag is True:
                    x = torch.add(_x, x)
                elif self.__hidden_dense_flag is True:
                    x = F.cat((_x, x), dim=self.__dense_axis)

                if self.__hidden_residual_flag is True or self.__hidden_dense_flag is True:
                    _x = x

        if self.output_nn is not None:
            x = self.output_nn(x)

        return x

    def compute_acc(self, prob_arr, batch_target_arr):
        '''
        Compute accuracy.

        Args:
            prob_arr:               Softmax probabilities.
            batch_target_arr:       t-hot vectors.
        
        Returns:
            Tuple data.
            - Accuracy.
            - inferenced label.
            - real label.
        '''
        inferenced_label_arr = prob_arr.argmax(axis=1)
        answer_label_arr = batch_target_arr.argmax(axis=1)
        acc = (inferenced_label_arr == answer_label_arr).sum() / self.batch_size
        return acc.to('cpu').detach().numpy(), inferenced_label_arr, answer_label_arr

    def save_parameters(self, filename):
        '''
        Save parameters to files.

        Args:
            filename:       File name.
        '''
        torch.save(
            {
                'epoch': self.epoch,
                'model_state_dict': self.state_dict(),
                'optimizer_state_dict': self.optimizer.state_dict(),
                'loss': self.loss_arr,
            }, 
            filename
        )

    def load_parameters(self, filename, ctx=None, strict=True):
        '''
        Load parameters to files.

        Args:
            filename:       File name.
            ctx:            Context-manager that changes the selected device.
            strict:         Whether to strictly enforce that the keys in state_dict match the keys returned by this module’s state_dict() function. Default: `True`.
        '''
        checkpoint = torch.load(filename)
        self.load_state_dict(checkpoint['model_state_dict'], strict=strict)
        self.optimizer.load_state_dict(
            checkpoint['optimizer_state_dict']
        )
        self.epoch = checkpoint['epoch']
        self.__loss_list = checkpoint['loss'].tolist()
        if ctx is not None:
            self.to(ctx)
            self.__ctx = ctx

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def set_batch_size(self, value):
        ''' setter for batch size. '''
        self.__batch_size = value

    def get_batch_size(self):
        ''' getter for batch size. '''
        return self.__batch_size
    
    batch_size = property(get_batch_size, set_batch_size)

    __loss_list = []

    def get_loss_arr(self):
        ''' getter for for `list` of accuracies. '''
        return np.array(self.__loss_list)

    loss_arr = property(get_loss_arr, set_readonly)

    def get_acc_list(self):
        ''' getter for `list` of accuracies. '''
        return np.array(self.__acc_list)
    
    acc_arr = property(get_acc_list, set_readonly)

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not.'''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not. '''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)
